/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.communication.jndi.client;

import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.communication.jndi.agent.JNDIAgentService;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.DomainConfiguration;
import org.ow2.petals.util.LoggingUtil;

/**
 * Fractal component used to retrieve the JNDI initial context.
 * 
 * @author alouis
 * @author Christophe HAMERLING - eBM WebSourcing
 * @author Roland NAUDIN - eBM WebSourcing
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.communication.jndi.client.JNDIService.class))
public class JNDIServiceImpl implements JNDIService {

    /**
     * Users context
     */
    protected Context usersContext;

    /**
     * Logging utilities.
     */
    protected LoggingUtil log;

    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * The JNDI Agent service fractal component
     */
    @Requires(name = "jndiAgent", signature = org.ow2.petals.communication.jndi.agent.JNDIAgentService.class)
    private JNDIAgentService jndiAgentService;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The domain configuration
     */
    private DomainConfiguration domainConfiguration;

    private InitialContextDecorator initialContextDecorator;

    @LifeCycle(on = LifeCycleType.START)
    protected void start() {
        log = new LoggingUtil(logger);
        log.call();

        domainConfiguration = configurationService.getDomainConfiguration();

    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        log.start();

        if (initialContextDecorator != null) {
            try {
                initialContextDecorator.closeInitialContextDecorator();
            } catch (NamingException e) {
                log.error(e);
            } finally {
                initialContextDecorator = null;
            }
        }

        log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.kernel.jndi.JNDIService#getJNDIContext()
     */
    public InitialContext getInitialContext() throws NamingException {
        log.start();

        if (initialContextDecorator == null) {
            log.debug("Create a new InitialContext decorator");
            Hashtable<String, Object> env = getJNDIEnv();
            log.debug("JNDI properties used : " + env);

            initialContextDecorator = new InitialContextDecorator(env, log, jndiAgentService);

        }

        log.end();
        return initialContextDecorator;
    }

    public Context getUsersContext() throws NamingException {
        if (usersContext == null) {
            try {
                usersContext = (Context) getInitialContext().lookup(USERS_REF);
            } catch (NameNotFoundException e) {
                usersContext = getInitialContext().createSubcontext(USERS_REF);
            } catch (NameAlreadyBoundException e) {
                // in distributed mode, it is possible that 2 containers are
                // creating the context in the same time, the second
                // container get an Already Bound exception
                usersContext = (Context) getInitialContext().lookup(USERS_REF);
            }
        }
        return usersContext;
    }

    /**
     * Construct and return the JNDI environment values. If no JNDI server is
     * specified in the Domain topology, use the internal JNDI server
     * 
     * @return The environment hashTable
     */
    private Hashtable<String, Object> getJNDIEnv() {
        Hashtable<String, Object> env = new Hashtable<String, Object>();

        if (domainConfiguration.getJndiFactory() == null) {
            env.put(JNDIAgentService.ENVIRONMENT_LOCAL_CONNECTION, Boolean.toString(true));
            env.put(Context.INITIAL_CONTEXT_FACTORY, JNDIAgentService.JNDI_PETALS_FACTORY);
        } else {
            env.put(Context.INITIAL_CONTEXT_FACTORY, domainConfiguration.getJndiFactory());
            env.put(Context.PROVIDER_URL, domainConfiguration.getJndiProviderUrl().toString());
            env.put(Context.SECURITY_PRINCIPAL, domainConfiguration.getJndiSecurityPrincipal());
            env.put(Context.SECURITY_CREDENTIALS, domainConfiguration.getJndiSecurityCredentials());
            env.put(Context.BATCHSIZE, String.valueOf(domainConfiguration.getJndiBatchSize()));
            env.put("java.naming.poolsize", String.valueOf(domainConfiguration.getJndiPoolSize()));
            env.put(JAVA_NAMING_FACTORY_LOGGER, this.log);
        }
        return env;
    }

}
