/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.communication.topology;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NameNotFoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.communication.network.NetworkService;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.configuration.DomainConfiguration;
import org.ow2.petals.kernel.configuration.SubDomainConfiguration;
import org.ow2.petals.kernel.configuration.DomainConfiguration.DomainMode;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.DreamService;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Jndi;
import org.ow2.petals.topology.generated.JoramService;
import org.ow2.petals.topology.generated.Multicast;
import org.ow2.petals.topology.generated.Subdomain;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.util.LoggingUtil;

/**
 * This {@link TopologyService} implementation is based on JNDI.
 * 
 * @author Roland NAUDIN - eBMWebSourcing
 * @since Petals 2.0
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.communication.topology.TopologyService.class))
public class TopologyServiceImpl implements TopologyService {

    /**
     * Name of the JNDI container context
     */
    private static final String TOPOLOGY_REF = "topology";

    private static final String DOMAIN_PREFIX = "domain_";

    private static final String SUBDOMAIN_PREFIX = "subdomain_";

    /**
     * Logging utilities
     */
    private LoggingUtil log;

    @Monolog(name = "logger")
    private Logger logger;

    /**
     * The JNDI service fractal component
     */
    @Requires(name = "jndi", signature = org.ow2.petals.communication.jndi.client.JNDIService.class)
    private JNDIService jndiService;

    /**
     * The Network service fractal component
     */
    @Requires(name = "network", signature = org.ow2.petals.communication.network.NetworkService.class)
    private NetworkService networkService;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The local container configuration
     */
    private ContainerConfiguration containerConfiguration;

    /**
     * The sub-domain configuration
     */
    private SubDomainConfiguration subdomainConfiguration;

    /**
     * The domain configuration
     */
    private DomainConfiguration domainConfiguration;

    private Context topologyContext;

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.communication.topology.TopologyService#
     * getContainersConfiguration
     * (org.ow2.petals.kernel.configuration.ContainerConfiguration
     * .ContainerState)
     */
    public Set<ContainerConfiguration> getContainersConfiguration(
            ContainerConfiguration.ContainerState state) throws PetalsException {
        this.log.start("Get containers configuration with state " + state);

        Set<ContainerConfiguration> containers = new HashSet<ContainerConfiguration>();
        try {
            InitialContext rootContext = this.jndiService.getInitialContext();
            NamingEnumeration<Binding> enums = rootContext.listBindings(TOPOLOGY_REF);
            while (enums.hasMoreElements()) {
                Binding binding = enums.next();
                if (binding.getClassName().equals(ContainerConfiguration.class.getName())) {
                    ContainerConfiguration container = (ContainerConfiguration) binding.getObject();

                    // get the entire configuration when it is the local
                    // container
                    if (container.getName().equals(this.containerConfiguration.getName())) {
                        container = this.containerConfiguration;
                    }

                    if (state == null) {
                        containers.add(container);
                    } else if (state.equals(container.getState())) {
                        containers.add(container);
                    }
                }
            }
        } catch (NamingException e) {
            throw new PetalsException("Failed to retrieve containers configuration");
        }

        this.log.end("Returns " + containers.size() + " containers");
        return containers;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.communication.topology.TopologyService#
     * getSubDomainsConfiguration()
     */
    public Set<SubDomainConfiguration> getSubDomainsConfiguration() throws PetalsException {
        this.log.start();

        Set<SubDomainConfiguration> subdomains = new HashSet<SubDomainConfiguration>();
        try {
            InitialContext rootContext = this.jndiService.getInitialContext();
            NamingEnumeration<Binding> enums = rootContext.listBindings(TOPOLOGY_REF);
            while (enums.hasMoreElements()) {
                Binding binding = enums.next();
                if (binding.getClassName().equals(SubDomainConfiguration.class.getName())) {
                    subdomains.add((SubDomainConfiguration) binding.getObject());
                }
            }
        } catch (NamingException e) {
            throw new PetalsException("Failed to retrieve sub-domains configuration");
        }

        this.log.end();
        return subdomains;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.communication.topology.TopologyService#
     * getContainerConfiguration(java.lang.String)
     */
    public ContainerConfiguration getContainerConfiguration(String containerName)
            throws PetalsException {
        this.log.start("Get container configuration : " + containerName);

        ContainerConfiguration conf = null;
        try {
            conf = (ContainerConfiguration) this.topologyContext.lookup(containerName);
        } catch (NamingException e) {
            throw new PetalsException("Container configuration has not been found for container : "
                    + containerName);
        }

        this.log.end("Returns : " + conf);
        return conf;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.communication.topology.TopologyService#setContainerState
     * (java.lang.String,
     * org.ow2.petals.kernel.configuration.ContainerConfiguration
     * .ContainerState)
     */
    public void setContainerState(String containerName, ContainerConfiguration.ContainerState state)
            throws PetalsException {
        this.log.start("Set state " + state + " to the container named " + containerName);

        if (containerName == null) {
            throw new IllegalArgumentException("The container name must not be null");
        }
        if (state == null) {
            throw new IllegalArgumentException("The state must not be null");
        }

        ContainerConfiguration conf = null;
        if (containerName.equals(this.containerConfiguration.getName())) {
            conf = this.containerConfiguration;
        } else {
            conf = this.getContainerConfiguration(containerName);
        }

        if (!conf.getState().equals(state)) {
            conf.setState(state);
            try {
                this.topologyContext.rebind(conf.getName(), conf);
            } catch (NamingException e) {
                throw new PetalsException("Failed to set a state for the container named '"
                        + containerName + "'");
            }
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.kernel.registry.ContainerNetwork#isContainerStarted
     * (java.lang.String)
     */
    public boolean isContainerStarted(String containerName) throws PetalsException {
        this.log.start("Container started? " + containerName);

        if (containerName == null) {
            throw new IllegalArgumentException("The container name must not be null");
        }

        boolean isStarted = false;
        ContainerConfiguration conf = this.getContainerConfiguration(containerName);
        isStarted = ContainerConfiguration.ContainerState.STARTED.equals(conf.getState());

        this.log.end("Returns : " + isStarted);
        return isStarted;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.communication.topology.TopologyService#unregisterContainer
     * (java.lang.String)
     */
    public void unregisterContainer(String containerName) throws PetalsException {
        this.log.start();

        try {
            this.topologyContext.unbind(this.containerConfiguration.getName());
        } catch (NamingException e) {
            throw new PetalsException("Failed to unregister container", e);
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.communication.topology.TopologyService#getDynamicTopology
     * ()
     */
    public Topology getDynamicTopology() throws PetalsException {
        this.log.start();

        if (!this.domainConfiguration.getMode().equals(DomainConfiguration.DomainMode.DYNAMIC)) {
            throw new PetalsException("The domain is not in dynamic mode");
        }

        Topology topology = new Topology();

        Multicast multicast = new Multicast();
        multicast.setIp(this.domainConfiguration.getNetworkMuticastIp());
        multicast.setPort(this.domainConfiguration.getNetworkMuticastPort());

        Jndi jndi = new Jndi();
        if (this.domainConfiguration.getJndiFactory() != null) {
            jndi.setFactory(this.domainConfiguration.getJndiFactory());
            jndi.setProviderUrl(this.domainConfiguration.getJndiProviderUrl().toString());
            jndi.setSecurityPrincipal(this.domainConfiguration.getJndiSecurityPrincipal());
            jndi.setSecurityCredentials(this.domainConfiguration.getJndiSecurityCredentials());
            jndi.setPoolsize(this.domainConfiguration.getJndiPoolSize());
            jndi.setBatchsize(this.domainConfiguration.getJndiBatchSize());
        }

        Domain domain = new Domain();
        domain.setName(this.domainConfiguration.getName());
        domain.setMode(org.ow2.petals.topology.generated.DomainMode.DYNAMIC);
        domain.setMulticast(multicast);
        domain.setJndi(jndi);
        topology.setDomain(domain);

        Set<SubDomainConfiguration> subdomainsConfiguration = this.getSubDomainsConfiguration();

        for (SubDomainConfiguration remoteSubdomain : subdomainsConfiguration) {
            Subdomain subdomain = new Subdomain();
            subdomain.setName(remoteSubdomain.getName());
            subdomain.setNetworkTimeSynchronized(remoteSubdomain.isNetworkTimeSynchronized());
            topology.getDomain().getSubDomain().add(subdomain);
        }

        Set<ContainerConfiguration> containersConfiguration = this.getContainersConfiguration(null);

        for (ContainerConfiguration remoteContainerConfiguration : containersConfiguration) {
            org.ow2.petals.topology.generated.NetworkService networkService = new org.ow2.petals.topology.generated.NetworkService();
            networkService.setPort(remoteContainerConfiguration.getNetworkBindingPort());

            JmxService jmxService = new JmxService();
            jmxService.setRmiPort(remoteContainerConfiguration.getJmxRMIConnectorPort());

            JoramService joramService = new JoramService();
            joramService.setDomainPort(remoteContainerConfiguration.getJoramDomainPort());
            joramService.setTcpPort(remoteContainerConfiguration.getJoramTCPPort());

            DreamService dreamService = new DreamService();
            dreamService.setTcpPort(remoteContainerConfiguration.getDreamTCPPort());
            dreamService.setSslPort(remoteContainerConfiguration.getDreamSSLPort());

            Container container = new Container();
            container.setName(remoteContainerConfiguration.getName());
            container.setHost(remoteContainerConfiguration.getHost());
            container.setUser(remoteContainerConfiguration.getUser());
            container.setPassword(remoteContainerConfiguration.getPassword());
            container.setNetworkService(networkService);
            container.setJmxService(jmxService);
            container.setJoramService(joramService);
            container.setDreamService(dreamService);
            for (Subdomain subdomain : topology.getDomain().getSubDomain()) {
                if (subdomain.getName().equals(remoteContainerConfiguration.getSubdomainName())) {
                    subdomain.getContainer().add(container);
                    break;
                }
            }
        }

        this.log.end();
        return topology;
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws PetalsException, TopologyException, NamingException {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        try {
            String response = this.networkService.joinNetwork();

            if (!response.equals(NetworkService.JOIN_ACCEPTED)) {
                throw new PetalsException(
                        "PEtALS network refused to accept the local container. Cause: " + response);
            }

            this.containerConfiguration = this.configurationService.getContainerConfiguration();
            this.subdomainConfiguration = this.configurationService.getSubDomainConfiguration();
            this.domainConfiguration = this.configurationService.getDomainConfiguration();

            this.getOrCreateTopologyContext();

            if (DomainMode.DYNAMIC.equals(this.domainConfiguration.getMode())) {
                Set<ContainerConfiguration> containers = this.getContainersConfiguration(null);
                this.verifyContainerDynamicTopology(containers);
            }

            this.registerLocalTopology();
        } catch (NamingException e) {
            this.log.error("Failed to start Topology service", e);
            throw e;
        } catch (PetalsException e) {
            this.log.error("Failed to start Topology service", e);
            throw e;
        } catch (TopologyException e) {
            this.log.error("Failed to start Topology service", e);
            throw e;
        }

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws Exception {
        this.log.call();
    }

    /**
     * Get or create the topology JNDI context.
     * 
     * @return The topology context
     * @throws PetalsException
     *             Failed to create or retrieve the topology JNDI context
     */
    private void getOrCreateTopologyContext() throws NamingException {
        this.log.start();

        InitialContext rootContext = this.jndiService.getInitialContext();
        try {
            this.topologyContext = (Context) rootContext.lookup(TOPOLOGY_REF);
        } catch (NameNotFoundException e) {
            this.topologyContext = rootContext.createSubcontext(TOPOLOGY_REF);
        }

        this.log.end();
    }

    /**
     * Verify the local container configuration against the dynamic topology.
     * 
     * @param containers
     *            The list of the domain' containers
     * @throws PetalsException
     */
    private void verifyContainerDynamicTopology(Set<ContainerConfiguration> containers)
            throws TopologyException, PetalsException {
        this.log.start();

        Collection<Integer> ports = new HashSet<Integer>();
        ports.add(this.containerConfiguration.getNetworkBindingPort());
        if (this.containerConfiguration.getJmxRMIConnectorPort() != 0) {
            ports.add(this.containerConfiguration.getJmxRMIConnectorPort());
        }
        ports.add(this.containerConfiguration.getJoramDomainPort());
        if (this.containerConfiguration.getJoramTCPPort() != 0) {
            ports.add(this.containerConfiguration.getJoramTCPPort());
        }
        ports.add(this.containerConfiguration.getDreamTCPPort());

        for (ContainerConfiguration remoteContainerConfiguration : containers) {
            if (remoteContainerConfiguration.getName()
                    .equals(this.containerConfiguration.getName())) {
                if (ContainerConfiguration.ContainerState.STARTED
                        .equals(remoteContainerConfiguration.getState())) {
                    throw new PetalsException("Container named '"
                            + this.containerConfiguration.getName() + "' is already started");
                }
                break;
            } else if (remoteContainerConfiguration.getHost().equals(
                    this.containerConfiguration.getHost())) {
                if (!ports.add(remoteContainerConfiguration.getNetworkBindingPort())) {
                    throw new TopologyException("Duplicate port "
                            + remoteContainerConfiguration.getNetworkBindingPort());
                }
                if (!ports.add(remoteContainerConfiguration.getJmxRMIConnectorPort())) {
                    throw new TopologyException("Duplicate port "
                            + remoteContainerConfiguration.getJmxRMIConnectorPort());
                }
                if (!ports.add(remoteContainerConfiguration.getJoramDomainPort())) {
                    throw new TopologyException("Duplicate port "
                            + remoteContainerConfiguration.getJoramDomainPort());
                }

                if (remoteContainerConfiguration.getJoramTCPPort() != 0
                        && !ports.add(remoteContainerConfiguration.getJoramTCPPort())) {
                    throw new TopologyException("Duplicate port "
                            + remoteContainerConfiguration.getJoramTCPPort());
                }
                if (remoteContainerConfiguration.getDreamTCPPort() != 0
                        && !ports.add(remoteContainerConfiguration.getDreamTCPPort())) {
                    throw new TopologyException("Duplicate port "
                            + remoteContainerConfiguration.getDreamTCPPort());
                }
                if (remoteContainerConfiguration.getDreamSSLPort() != 0
                        && !ports.add(remoteContainerConfiguration.getDreamSSLPort())) {
                    throw new TopologyException("Duplicate port "
                            + remoteContainerConfiguration.getDreamSSLPort());
                }
            }
        }

        this.log.end();
    }

    /**
     * Register the topology elements related to the local container.
     * 
     * @param topologyContext
     * @throws PetalsException
     */
    private void registerLocalTopology() throws PetalsException {
        this.log.start();

        try {
            // domain
            try {
                this.topologyContext.lookup(DOMAIN_PREFIX + this.domainConfiguration.getName());
            } catch (NameNotFoundException e1) {
                this.topologyContext.bind(DOMAIN_PREFIX + this.domainConfiguration.getName(),
                        this.domainConfiguration);
            }

            // sub-domain
            try {
                this.topologyContext.lookup(SUBDOMAIN_PREFIX
                        + this.subdomainConfiguration.getName());
            } catch (NameNotFoundException e1) {
                this.topologyContext.bind(SUBDOMAIN_PREFIX + this.subdomainConfiguration.getName(),
                        this.subdomainConfiguration);
            }

            // container
            this.topologyContext.rebind(this.containerConfiguration.getName(),
                    this.containerConfiguration);
        } catch (NamingException e) {
            throw new PetalsException("Error while binding container information", e);
        }

        this.log.end();
    }

}
