/**
 * PETALS: PETALS Services Platform Copyright (C) 2005 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: LifeCycle.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.container.lifecycle;

import javax.jbi.JBIException;
import javax.jbi.management.LifeCycleMBean;

import org.ow2.petals.util.LoggingUtil;

/**
 * Implementation of the <code>LifeCycle</code> mbean. This lifecycle mbean
 * sends notifications about changes of state.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
public abstract class LifeCycleAbstract implements LifeCycleMBean {

    protected Object activitySynchronizer = new Byte[0];

    protected boolean initialized;

    protected LoggingUtil log;

    protected long sequenceNumber;

    private String state = null;

    protected String previousState;

    protected String name;

    /**
     * Creates a new instance of {@link LifeCycleAbstract}
     * 
     */
    public LifeCycleAbstract() {
        super();
        state = UNKNOWN;
    }

    /**
     * 
     * @throws JBIException
     */
    public abstract void doInit() throws JBIException;

    /**
     * 
     * @throws JBIException
     */
    public abstract void doShutdown() throws JBIException;

    /**
     * 
     * @throws JBIException
     */
    public abstract void doStart() throws JBIException;

    /**
     * 
     * @throws JBIException
     */
    public abstract void doStop() throws JBIException;

    protected abstract String getJBIArtifactType();

    /**
     * @see javax.jbi.management.LifeCycleMBean#getCurrentState()
     */
    public String getCurrentState() {
        synchronized (this.state) {
            if (!(SHUTDOWN.equals(state) || STARTED.equals(state) || STOPPED.equals(state) || UNKNOWN
                    .equals(state))) {
                return UNKNOWN;
            }
        }
        return state;
    }

    /**
     * Set the <code>State</code> reference of the current state. An
     * <code>LifeCycleStateEventAbstract</code> is thrown.
     * 
     * @param state
     */
    protected void setState(String state) throws JBIException {
        synchronized (this.state) {
            if (!(SHUTDOWN.equals(state) || STARTED.equals(state) || STOPPED.equals(state) || UNKNOWN
                    .equals(state))) {
                throw new JBIException("State '" + state
                        + "' isn't defined by the JBI specification. No state changement done.");
            }
            this.previousState = new String(this.state);
            this.state = state;
        }
    }

    /**
     * Shutdown the component. A notification will be thrown in case of
     * runtimeException or incoherent state.
     * 
     * @see javax.jbi.management.LifeCycleMBean#shutdown()
     * @throws JBIException
     *             in case of runtimeException or incoherent state.
     */
    public void shutDown() throws JBIException {
        log.start();

        if (SHUTDOWN.equals(state)) {
            log.warning("The component '" + this.name + "' is already shut down");
        } else if (STOPPED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    doShutdown();
                    setState(SHUTDOWN);
                    initialized = false;
                } catch (Throwable re) {
                    log.error(re.getMessage(), re);
                    setState(UNKNOWN);
                    JBIException e = new JBIException(
                            "An exception occured while shutting down the " + getJBIArtifactType()
                                    + " '" + this.name + "' : " + re.getMessage());
                    throw e;
                }
            }
        } else {
            JBIException e = new JBIException("The " + getJBIArtifactType() + " '" + this.name
                    + "' cannot be shut down in this state: " + state);
            throw e;
        }

        log.end();
    }

    /**
     * Start the component. A notification will be thrown in case of
     * runtimeException or incoherent state.
     * 
     * @see javax.jbi.management.LifeCycleMBean#start()
     * @throws JBIException
     *             in case of runtimeException or incoherent state.
     */
    public void start() throws JBIException {
        log.start();

        if (STARTED.equals(state)) {
            log.warning("The component '" + this.name + "' is already started");
        } else if (SHUTDOWN.equals(state) || STOPPED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    if (!initialized) {
                        doInit();
                        initialized = true;
                    }
                    doStart();
                    setState(STARTED);
                } catch (Throwable re) {
                    log.error(re.getMessage(), re);
                    setState(UNKNOWN);
                    JBIException e = new JBIException("An exception occured while starting the "
                            + getJBIArtifactType() + " '" + this.name + "' : " + re.getMessage());
                    throw e;
                }
            }
        } else {
            JBIException e = new JBIException("The " + getJBIArtifactType() + " '" + this.name
                    + "' cannot be started in this state: " + state);
            throw e;
        }

        log.end();
    }

    /**
     * Stop the component. A notification will be thrown in case of
     * runtimeException or incoherent state.
     * 
     * @see javax.jbi.management.LifeCycleMBean#stop()
     * @throws JBIException
     *             in case of runtimeException or incoherent state.
     */
    public void stop() throws JBIException {
        log.start();
        if (STOPPED.equals(state)) {
            log.warning("The component '" + this.name + "' is already stopped");
        } else if (STARTED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    doStop();
                    setState(STOPPED);
                } catch (Throwable re) {
                    log.error(re.getMessage(), re);
                    setState(UNKNOWN);
                    JBIException e = new JBIException("An exception occured while stopping the "
                            + getJBIArtifactType() + " '" + this.name + "' : " + re.getMessage());
                    throw e;
                }
            }
        } else {
            JBIException e = new JBIException("The " + getJBIArtifactType() + " '" + this.name
                    + "' cannot be stopped in this state: " + state);
            throw e;
        }
        log.end();
    }
}
