/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: UndeployAllSUsTask.java 154 10 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task.deployment.undeploy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;
import javax.jbi.management.LifeCycleMBean;

import org.ow2.petals.container.lifecycle.ComponentLifeCycle;
import org.ow2.petals.container.lifecycle.Installer;
import org.ow2.petals.container.lifecycle.ServiceAssemblyLifeCycle;
import org.ow2.petals.container.lifecycle.ServiceUnitLifeCycle;
import org.ow2.petals.jbi.management.admin.AdminService;
import org.ow2.petals.jbi.management.task.AbstractLoggableTask;
import org.ow2.petals.jbi.management.task.deployment.DeploymentContextConstants;
import org.ow2.petals.jbi.management.util.XMLResult;
import org.ow2.petals.util.LoggingUtil;

/**
 * This task executes the undeployment process of a service unit, for each
 * Service Unit included into the service assembly <br/>
 * Inputs : {@link ServiceAssemblyLifeCycle}, {@link XMLResult}<br />
 * The task has got a mixed approach: it tries to do all the subtasks<br />
 * and doesn't have the undo capability. But if the service assembly isn't over
 * shutdown or unknown state,<br />
 * to implement the JBI specification, it stops and thrown an exception to
 * launch the undo process
 * 
 * @author ofabre - EBM Websourcing
 * @author Frederic Gardes - eBM Websourcing
 */
public class UndeployAllSUsTask extends AbstractLoggableTask {

    /**
     * JBI Container :: JMX Agent Admin
     */
    protected AdminService adminService;

    public UndeployAllSUsTask(LoggingUtil log, AdminService managerService) {
        super(log);
        this.adminService = managerService;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#execute(java.util.HashMap)
     */
    public void execute(HashMap<String, Object> context) throws Exception {
        if (context.containsKey(DeploymentContextConstants.SA_LIFECYCLE)) {
            ServiceAssemblyLifeCycle saLifeCycle = (ServiceAssemblyLifeCycle) context
                    .get(DeploymentContextConstants.SA_LIFECYCLE);
            if (saLifeCycle != null && context.containsKey(DeploymentContextConstants.XML_RESULT)) {
                XMLResult xmlResult = (XMLResult) context
                        .get(DeploymentContextConstants.XML_RESULT);
                if (xmlResult != null) {
                    // Undeploy SUs
                    undeploySUS(saLifeCycle, xmlResult);
                }
            }
        }
    }

    /**
     * Undeploy all SUs by calling the "undeploySU" method of the corresponding
     * SUDeploymentManager.
     * 
     * @param saLifeCycle
     *            the parent {@link ServiceAssemblyLifeCycle}
     * @param xmlResult
     *            xml task result
     * @throws Exception
     *             if an error occurs during the undeployment of a service unit
     */
    protected void undeploySUS(ServiceAssemblyLifeCycle saLifeCycle, XMLResult xmlResult)
            throws JBIException {
        /*
         * Undeploy the SUs only if the SA is in SHUTDOWN or UNKNOW state
         */
        String saState = saLifeCycle.getCurrentState();
        if (saState != null
                && (LifeCycleMBean.SHUTDOWN.equals(saState) || LifeCycleMBean.UNKNOWN
                        .equals(saState))) {
            /*
             * JBI spec specifies that service units must be undeployed in the
             * reverse order than the order of the initial declared list
             */
            List<ServiceUnitLifeCycle> suLifeCycles = new ArrayList<ServiceUnitLifeCycle>(
                    saLifeCycle.getServiceUnitLifeCycles());
            Collections.reverse(suLifeCycles);
            StringBuffer componentOutput = new StringBuffer();
            for (ServiceUnitLifeCycle suLifeCycle : suLifeCycles) {
                String xmlReturn = undeploySU(suLifeCycle);
                //unregister the SU from the SA
                saLifeCycle.unregisterSU(suLifeCycle.getSuName());
                componentOutput.append(xmlReturn);
            }
            xmlResult.addComponentTaskResult(componentOutput);
        } else {
            throw new JBIException("Service Assembly can't be undeployed in this state: " + saState);
        }
    }

    /**
     * Process the undeployment of the given service unit.
     * 
     * @param suLifeCycle
     *            the service unit life cycle
     * @throws Exception
     *             if an error occurs during the undeployment process : can't
     *             retrieve the {@link ComponentLifeCycle}, can't retrieve the
     *             component {@link ServiceUnitManager} (SUM) or an exception is
     *             thrown during the call of the undeploy method of the SUM
     */
    protected String undeploySU(ServiceUnitLifeCycle suLifeCycle) {
        String xmlReturn = null;

        // Retrieve the component life cycle
        Installer installer = adminService.getInstallerByName(suLifeCycle.getTargetComponentName());
        if (installer == null) {
            log.error("The component '" + suLifeCycle.getTargetComponentName()
                    + "' is not deployed. The service unit cannot be undeployed from it.");
        } else {
            // Retrieve the service unit manager of the component
            ServiceUnitManager sum = installer.getComponent().getServiceUnitManager();
            if (sum == null) {
                log.error("The service unit manager is null for the component '"
                        + suLifeCycle.getTargetComponentName()
                        + "'. The service units can not be undeployed from it.");
            } else {
                // Set the context class loader to the component's one just
                // for the
                // method invocation
                try {
                    Thread.currentThread().setContextClassLoader(sum.getClass().getClassLoader());
                    xmlReturn = sum.undeploy(suLifeCycle.getSuName(), suLifeCycle
                            .getServiceUnitRootPath());
                } catch (DeploymentException e) {
                    log.error("The service unit manager of the component '"
                            + suLifeCycle.getTargetComponentName()
                            + "' cannot  undeploy a service unit.");
                } finally {
                    Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
                }
            }
        }

        return xmlReturn;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#undo(java.util.HashMap)
     */
    public void undo(HashMap<String, Object> context) throws Exception {
        // TODO : redeploy all the service units
    }

}
