/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task.installation.install.component;

import java.util.HashMap;

import javax.jbi.JBIException;
import javax.management.ObjectName;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.ow2.petals.container.ContainerService;
import org.ow2.petals.container.lifecycle.Installer;
import org.ow2.petals.container.lifecycle.InstallerInterface;
import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.management.ManagementException;
import org.ow2.petals.jbi.management.admin.AdminService;
import org.ow2.petals.jbi.management.task.AbstractLoggableTask;
import org.ow2.petals.jbi.management.task.TaskContextConstants;
import org.ow2.petals.jbi.management.task.installation.InstallationContextConstants;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.server.FractalHelper;
import org.ow2.petals.util.LoggingUtil;

/**
 * This task creates the InstallerMBean for a component and register it in the
 * JMX server <br/>
 * The task has got a non-robust approach : it stops at the first subtask failed<br />
 * and have the undo capability.<br />
 * Inputs : {@link Jbi}, component install root URI,
 * {@link ComponentContextImpl} <br/>
 * Outputs : component installer {@link ObjectName} <br/>
 * 
 * @author ofabre
 * @author Frederic Gardes - eBM Websourcing
 */
public class CreateAndRegisterInstallerMBeanTask extends AbstractLoggableTask {
    /**
     * JBI Container :: JMX Agent Admin
     */
    private final AdminService adminService;

    protected ContainerService containerService;

    /**
     * Creates a new instance of {@link CreateAndRegisterInstallerMBeanTask}
     * 
     * @param log
     * @param adminService
     * @param logger
     * @param router
     * @param containerService
     */
    public CreateAndRegisterInstallerMBeanTask(LoggingUtil log, AdminService adminService,
            ContainerService container) {
        super(log);
        this.adminService = adminService;
        this.containerService = container;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#execute(java.util.HashMap)
     */
    public void execute(HashMap<String, Object> context) throws Exception {
        final ObjectName result = this.createAndRegisterInstallerMBean(context);
        context.put(InstallationContextConstants.INSTALLER_OBJECT_NAME, result);
    }

    /**
     * Create and register a new InstallerMBean for the loaded component
     * 
     * @param componentContext
     *            the new Component execution context
     * @param descriptor
     *            the component Jbi
     * @param installationRoot
     *            the URI representing the component installation folder
     * @return the new InstallerMBean name
     * @throws PetalsException
     */
    private ObjectName createAndRegisterInstallerMBean(HashMap<String, Object> ctxt)
            throws PetalsException {
        ObjectName result = null;
        try {
            final Jbi descriptor = (Jbi) ctxt.get(TaskContextConstants.DESCRIPTOR);
            final org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription = descriptor
                    .getComponent();
            // Create the Fractal installer
            final Component installer = this.containerService.createInstaller(componentDescription);
            // Start the Fractal installer
            FractalHelper.startComponent(installer);
            // Initialize the Fractal installer
            ((InstallerInterface) installer.getFcInterface("service")).init(ctxt);
            // Register the InstallerMBean
            result = this.adminService
                    .registerInstaller((Installer) installer.getFcInterface("/content"),
                            componentDescription.getIdentification().getName());
        } catch (final ManagementException e) {
            throw new PetalsException("Error creating or registring installerMBean", e);
        } catch (final NoSuchInterfaceException e) {
            throw new PetalsException("Error creating or registring installerMBean", e);
        } catch (final IllegalLifeCycleException e) {
            throw new PetalsException("Error creating or registring installerMBean", e);
        } catch (final JBIException e) {
            throw new PetalsException("Error creating or registring installerMBean", e);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#undo(java.util.HashMap)
     */
    public void undo(HashMap<String, Object> context) throws Exception {
        final Jbi descriptor = (Jbi) context.get(TaskContextConstants.DESCRIPTOR);
        final String componentName = descriptor.getComponent().getIdentification().getName();

        try {
            // Unregister the InstallerMBean
            this.adminService.unregisterInstaller(componentName);
        } catch (final ManagementException e) {
            this.log.error("Component installer can't be unregistered or doesn't exist", e);
        }

        try {
            containerService.removeInstaller(componentName);
        } catch (PetalsException e) {
            log.error("Fractal installer can't be removed or doesn't exist", e);
        }
    }
}
