/**
 * PETALS: PETALS Services Platform Copyright (C) 2005 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: DeliveryChannelImpl.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.exchange;

import javax.jbi.JBIException;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.transaction.SystemException;
import javax.xml.namespace.QName;

import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.management.transaction.TransactionManagerImpl;
import org.ow2.petals.jbi.messaging.routing.RoutingException;
import org.ow2.petals.util.LoggingUtil;

/**
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Gael BLONDELLE - EBM WebSourcing
 * @author ofabre - EBM Websourcing
 * @author rnaudin - EBM Websourcing
 */
public class DeliveryChannelImpl implements DeliveryChannel {

    /**
     * The intern message exchange property name to mark the exchange as a
     * synchronous invocation.
     */
    public static final String PROPERTY_SENDSYNC = "javax.jbi.messaging.sendSync";

    public static final String PROPERTY_OVERSENDSYNC = "javax.jbi.messaging.overSendSync";

    protected ComponentContextImpl context;

    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    protected Logger logger;

    /**
     * This messageExchangeFactory is used to return
     * <code>MessageExchangeDecorator</code> on accept() methods.
     */
    protected MessageExchangeFactoryImpl messageExchangeFactory;

    protected boolean opened = true;

    /**
     * Constructor
     * 
     * @param context
     * @param router
     * @param logger
     */
    public DeliveryChannelImpl(ComponentContextImpl context) {
        this.logger = context.getLogger();

        this.log = new LoggingUtil(this.logger);

        this.messageExchangeFactory = new MessageExchangeFactoryImpl(this, context.getAddress(),
                this.logger);

        this.context = context;
    }

    /**
     * Return a MessageExchangeDecorator with 'consumer' or 'provider' observer
     * role (depending on the messageExchange getRole())
     * 
     * @see javax.jbi.messaging.DeliveryChannel#accept()
     */
    public MessageExchange accept() throws MessagingException {
        this.log.call();
        MessageExchange me = this.poll(-1);

        return me;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#accept(long)
     */
    public MessageExchange accept(final long timeoutMS) throws MessagingException {
        this.log.call(timeoutMS);
        if (timeoutMS < 0) {
            throw new MessagingException("Timeout must be positive or null");
        }
        MessageExchange me = this.poll(timeoutMS);
        return me;
    }

    /**
     * Check if the <code>DeliveryChannel</code> is opened
     * 
     * @throws MessagingException
     */
    public synchronized void checkDeliveryChannelIsOpened() throws MessagingException {
        if (!this.opened) {
            throw new MessagingException("DeliveryChannel is closed.");
        }
    }

    /**
     * Check if the <code>MessageExchange</code> is Ok (not null and not
     * terminated)
     * 
     * @throws MessagingException
     */
    public void checkMessageExchange(MessageExchange exchange) throws MessagingException {
        if (exchange == null) {
            throw new MessagingException("MessageExchange is null.");
        }
        MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;

        if (exchangeDecorator.getMessageExchange().isTerminated()) {
            throw new MessagingException("MessageExchange is terminated.");
        }
    }

    /**
     * This method wake up all threads that are waiting for a message. This
     * method deregister all endpoints for this component.
     * 
     * @see javax.jbi.messaging.DeliveryChannel#close()
     */
    public synchronized void close() throws MessagingException {
        this.log.start();

        this.checkDeliveryChannelIsOpened();

        try {
            this.context.deregisterAllEndpoints();
        } catch (JBIException e) {
            throw new MessagingException(e);
        }

        this.opened = false;
        this.log.end();
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory()
     */
    public MessageExchangeFactory createExchangeFactory() {
        this.log.call();

        MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this, this.context
                .getAddress(), this.logger);

        return mef;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory(javax.xml.
     *      namespace.QName)
     */
    public MessageExchangeFactory createExchangeFactory(QName interfaceName) {
        this.log.call(interfaceName);

        MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this, this.context
                .getAddress(), this.logger);

        mef.setDefaultInterfaceName(interfaceName);

        return mef;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactory(javax.jbi.
     *      servicedesc.ServiceEndpoint)
     */
    public MessageExchangeFactory createExchangeFactory(ServiceEndpoint endpoint) {
        this.log.call(endpoint);

        MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this, this.context
                .getAddress(), this.logger);

        mef.setDefaultServiceEndpoint(endpoint);

        return mef;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#createExchangeFactoryForService(javax.xml.namespace.QName)
     */
    public MessageExchangeFactory createExchangeFactoryForService(QName serviceName) {
        this.log.call(serviceName);

        MessageExchangeFactoryImpl mef = new MessageExchangeFactoryImpl(this, this.context
                .getAddress(), this.logger);

        mef.setDefaultServiceName(serviceName);

        return mef;
    }

    public synchronized boolean isOpened() {
        return this.opened;
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#send(javax.jbi.messaging.
     *      MessageExchange)
     */
    public void send(MessageExchange exchange) throws MessagingException {
        this.log.call("Exchange Id:" + exchange.getExchangeId());
        
        this.sendExchange(exchange, 0, false);
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#sendSynch(javax.jbi.messaging.
     *      MessageExchange)
     */
    public boolean sendSync(MessageExchange exchange) throws MessagingException {
        return this.sendSync(exchange, 0);
    }

    /**
     * @see javax.jbi.messaging.DeliveryChannel#sendSync(javax.jbi.messaging.
     *      MessageExchange, long)
     */
    public boolean sendSync(MessageExchange exchange, long timeoutMS) throws MessagingException {
        this.log.call("Exchange Id:" + exchange.getExchangeId() + " - Timeout:" + timeoutMS);

        MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;

        try {
            // prepare the send
            exchangeDecorator.setWaitingOnSynchronousSend(true);

            MessageExchangeImpl responseExchange = this.sendExchange(exchangeDecorator, timeoutMS,
                    true);

            if (responseExchange != null) {
                exchangeDecorator.setMessageExchange(responseExchange);
                exchangeDecorator.setWaitingOnSynchronousSend(false);
            } else {
                // get the ownership of the ME
                exchangeDecorator.getMessageExchange().setRole(exchangeDecorator.getObserverRole());
                exchangeDecorator.setStatus(ExchangeStatus.ERROR);
                exchangeDecorator.getMessageExchange().setTerminated(true);
            }

        } catch (MessagingException e) {
            this.log.error(e.getMessage(), e);
            // get the ownership of the ME
            exchangeDecorator.getMessageExchange().setRole(exchangeDecorator.getObserverRole());
            exchangeDecorator.setStatus(ExchangeStatus.ERROR);
            exchangeDecorator.getMessageExchange().setTerminated(true);

            // If the cause is an Interrupted exception, we return false
            Throwable causeException = e.getCause();
            while (causeException != null) {
                if (causeException instanceof InterruptedException) {
                    this.log.end();
                    return false;
                } else {
                    causeException = causeException.getCause();
                }
            }
            throw e;
        }

        this.log.end();

        return !exchangeDecorator.isWaitingOnSynchronousSend();
    }

    /**
     * Send the exchange to the underlying layer.<br>
     * Unused data in the exchange are cleaned before.<br>
     * If the message to exchange is a mere status information, and the exchange
     * is set to not send acknowledgment, the message is not sent.<br>
     */
    private MessageExchangeImpl sendExchange(MessageExchange exchange, long timeOut, boolean sync)
            throws MessagingException {
        this.checkDeliveryChannelIsOpened();
        this.checkMessageExchange(exchange);

        MessageExchangeDecorator exchangeDecorator = (MessageExchangeDecorator) exchange;
        MessageExchangeImpl exchangeImpl = exchangeDecorator.getMessageExchange();
        MessageExchangeImpl responseExchange = null;

        try {
            // check if the exchange is terminated
            if (!ExchangeStatus.ACTIVE.equals(exchangeImpl.getStatus())) {
                exchangeImpl.setTerminated(true);
                if (sync) {
                    throw new MessagingException(
                            "Sending an acknowledgment in synchronous mode is forbidden");
                }
            }

            // encode the transaction to the Message Exchange if it is
            // transacted
            if (this.context.getTransactionManager() != null) {
                TransactionManagerImpl.encodeTransactionContext(exchangeImpl);
            }

            if (sync) {
                responseExchange = this.context.getRouter().sendSync(this.context, exchangeImpl,
                        timeOut);
                if (responseExchange != null && this.context.getTransactionManager() != null) {
                    // decode the transaction to the Message Exchange if it is
                    // transacted
                    TransactionManagerImpl.decodeTransactionContext(responseExchange);
                }
            } else {
                this.context.getRouter().send(this.context, exchangeImpl, timeOut);
            }

        } catch (JBIException e) {
            throw new MessagingException(e);
        } catch (SystemException e) {
            throw new MessagingException(e);
        }

        this.log.end();
        return responseExchange;
    }

    /**
     * Poll for a receipt message exchange.
     * 
     * @param timeoutMS
     * 
     * @return The received message exchange
     * 
     * @throws MessagingException
     */
    private MessageExchange poll(long timeoutMS) throws MessagingException {
        this.log.start();

        this.checkDeliveryChannelIsOpened();

        MessageExchangeImpl msg = null;

        try {
            msg = this.context.getRouter().receive(this.context, timeoutMS);
        } catch (RoutingException e) {
            throw new MessagingException(e);
        }

        MessageExchange result = null;
        if (msg != null) {
            result = this.messageExchangeFactory.createExchangeDecorator(msg, msg.getPattern());
            // decode the transaction to the Message Exchange if it is
            // transacted
            if (this.context.getTransactionManager() != null) {
                TransactionManagerImpl.decodeTransactionContext(((MessageExchangeDecorator) result)
                        .getMessageExchange());
            }
        }

        this.log.end();

        return result;
    }

}
