/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.routing;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.jbi.messaging.routing.module.RouterModule;
import org.ow2.petals.jbi.messaging.routing.module.RouterModuleLoader;
import org.ow2.petals.jbi.messaging.routing.module.RouterModuleManagerMBean;

/**
 * The global router module contains all the router modules. It is in charge of
 * calling all the modules in the good order.
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 2.2
 * 
 */
// @FractalComponent
// @Provides(interfaces = @Interface(name = "service", signature =
// RouterModuleManagerMBean.class))
public class GlobalRouterModule extends RouterModule implements RouterModuleManagerMBean {

    /**
     * List of all the modules to call. Modules are ordered in the order to be
     * called. TODO : Add JMX Bean to add/start/stop modules at runtime.
     */
    private List<RouterModule> modules;

    /**
     * Creates a new instance of GlobalRouterModule
     * 
     * @param modules
     */
    public GlobalRouterModule() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.RouterModule#init()
     */
    @Override
    protected void doInit() throws Exception {
        this.loadModulesFromConfiguration();

        if (this.modules == null || this.modules.size() == 0) {
            throw new Exception(
                    "No router module has been loaded, please check your server configuration file");
        }

        for (int i = 0; i < this.modules.size(); i++) {
            RouterModule rm = this.modules.get(i);
            log.debug("Initialize the module : " + rm.getName());
            rm.init(this.routerService);
            rm.init((Map<String, String>) null);
        }
    }

    /**
     * Load the modules
     * 
     * @param router
     * @throws Exception
     */
    private void loadModulesFromConfiguration() {
        RouterModuleLoader loader = new RouterModuleLoader(logger);
        loader.load();
        this.modules = loader.getModules();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.module.RouterModule#doStart()
     */
    @Override
    protected void doStart() throws Exception {
        for (int i = 0; i < this.modules.size(); i++) {
            RouterModule rm = this.modules.get(i);
            log.debug("Starting the module : " + rm.getName());
            rm.start();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.jbi.messaging.routing.module.RouterModule#doStop()
     */
    @Override
    protected void doStop() throws Exception {
        for (int i = 0; i < this.modules.size(); i++) {
            RouterModule rm = this.modules.get(i);
            log.debug("Stopping the module : " + rm.getName());
            rm.stop();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.RouterModule#addComponent(org.ow2
     * .petals.jbi.component.context.ComponentContextImpl)
     */
    @Override
    public void addComponent(ComponentContextImpl componentContext) throws RoutingException {
        for (Iterator<RouterModule> iterator = this.modules.iterator(); iterator.hasNext();) {
            RouterModule routerModule = iterator.next();
            log.debug("Call addComponent for module " + routerModule.getName());
            routerModule.addComponent(componentContext);

        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.RouterModule#receive(org.ow2.petals
     * .jbi.component.context.ComponentContextImpl, long)
     */
    @Override
    public MessageExchangeImpl receive(ComponentContextImpl source, long timeoutMS)
            throws RoutingException {
        MessageExchangeImpl result = null;
        for (Iterator<RouterModule> iterator = this.modules.iterator(); iterator.hasNext();) {
            RouterModule routerModule = iterator.next();
            log.debug("Call receive for module " + routerModule.getName());

            MessageExchangeImpl tmp = routerModule.receive(source, timeoutMS);
            if (tmp != null) {
                // cache the last non null receive...
                result = tmp;
            }
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.RouterModule#removeComponent(org
     * .ow2.petals.jbi.component.context.ComponentContextImpl)
     */
    @Override
    public void removeComponent(ComponentContextImpl componentContext) throws RoutingException {
        for (Iterator<RouterModule> iterator = this.modules.iterator(); iterator.hasNext();) {
            RouterModule routerModule = iterator.next();
            log.debug("Call remove component for module " + routerModule.getName());

            routerModule.removeComponent(componentContext);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.RouterModule#send(org.ow2.petals
     * .jbi.component.context.ComponentContextImpl,
     * org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, long)
     */
    @Override
    public void send(ComponentContextImpl source, MessageExchangeImpl exchange, long timeOut)
            throws RoutingException {
        for (Iterator<RouterModule> iterator = this.modules.iterator(); iterator.hasNext();) {
            RouterModule routerModule = iterator.next();
            log.debug("Call send for module " + routerModule.getName());

            routerModule.send(source, exchange, timeOut);

        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.RouterModule#sendSync(org.ow2.petals
     * .jbi.component.context.ComponentContextImpl,
     * org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, long)
     */
    @Override
    public MessageExchangeImpl sendSync(ComponentContextImpl source, MessageExchangeImpl exchange,
            long timeOut) throws RoutingException {
        MessageExchangeImpl result = exchange;
        boolean isNull;

        for (RouterModule routerModule : this.modules) {
            log.debug("Call sendSync for module " + routerModule.getName());

            MessageExchangeImpl tmp = routerModule.sendSync(source, result, timeOut);
            if (tmp != null) {
                isNull = false;
                result = tmp;
            } else {
                isNull = true;
            }

            if (isNull) {
                result = null;
            }
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.module.RouterModuleManagerMBean#
     * addModule(java.lang.String)
     */
    public String addModule(String className) {
        log.info("Try to add the module : " + className);
        return "This operation is not supported";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.module.RouterModuleManagerMBean#
     * getModules(java.lang.String)
     */
    public String[] getModules(String state) {
        String[] result = null;
        List<RouterModule> tmp = null;

        if (state == null || state.length() == 0) {
            tmp = this.modules;
        } else {
            tmp = new LinkedList<RouterModule>();
            for (RouterModule module : this.modules) {
                if (module.getStatus().toString().equalsIgnoreCase(state)) {
                    tmp.add(module);
                }
            }
        }
        result = new String[tmp.size()];
        for (int i = 0; i < tmp.size(); i++) {
            result[i] = tmp.get(i).getName();
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.module.RouterModuleManagerMBean#
     * initModule(java.lang.String)
     */
    public String initModule(String moduleName) throws Exception {
        log.info("Try to initialize the module : " + moduleName);
        return "This operation is not supported";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.module.RouterModuleManagerMBean#
     * startModule(java.lang.String)
     */
    public String startModule(String moduleName) throws Exception {
        log.info("Try to start the module : " + moduleName);
        return "This operation is not supported";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.routing.module.RouterModuleManagerMBean#
     * stopModule(java.lang.String)
     */
    public String stopModule(String moduleName) throws Exception {
        log.info("Try to stop the module : " + moduleName);
        return "This operation is not supported";
    }
}
