/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: RouterService.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.routing;

import java.util.Map;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.jbi.messaging.registry.EndpointRegistry;
import org.ow2.petals.jbi.messaging.routing.module.LifeCycleException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.util.LoggingUtil;

/**
 * Routes messages to their destinations. Performs an itinerary resolution,
 * adding for example a transformation service before delivering the message to
 * its recipient, and perform an address resolution, finding the endpoint
 * satisfying the the destination information (finding an endpoint for a
 * specified service). Resolutions are based on defined rules.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Christophe HAMERLING - eBM WebSourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.jbi.messaging.routing.Router.class))
public class RouterServiceImpl implements Router {

    /**
     * Logger wrapper.
     */
    protected LoggingUtil log;

    /**
     * The logger.
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * the endpoint service.
     */
    @Requires(name = "endpoint", signature = org.ow2.petals.jbi.messaging.registry.EndpointRegistry.class)
    protected EndpointRegistry endpointService;

    /**
     * The Topology service fractal component
     */
    @Requires(name = "topology", signature = org.ow2.petals.communication.topology.TopologyService.class, contingency = Contingency.OPTIONAL)
    protected TopologyService topologyService;

    /**
     * the Transporter service fractal component
     */
    @Requires(name = "transporter", signature = org.ow2.petals.transport.Transporter.class)
    protected Transporter transporter;

    /**
     * The Topology service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    protected ConfigurationService configurationService;

    /**
     * The global router
     */
    protected GlobalRouterModule routerModule;

    /**
     * The default contructor.
     * 
     */
    public RouterServiceImpl() {
    }

    /**
     * start the router.
     * 
     * @throws RoutingException
     *             impossible to start the router
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws RoutingException {
        this.log = new LoggingUtil(this.logger);
        this.log.start();
        
        //
//        List<AuthorizationEntry> entries = new ArrayList<AuthorizationEntry>();
//        entries.add(new AuthorizationEntry(new QName("http://petals.ow2.org", "HelloService"),
//                null, null, "admin"));
//        this.authMap = new DefaultAuthorizationMap(entries);

        this.getRouterModule();
        
        // start the global router module
        try {
            this.routerModule.start();
        } catch (LifeCycleException e) {
            this.log.warning("Can not start the global router module", e);
            throw new RoutingException(e);
        }
        
        this.log.end();
    }

    /**
     * stop the router.
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws RoutingException {
        this.log.call();
        
        try {
            this.routerModule.stop();
        } catch (LifeCycleException e) {
           throw new RoutingException(e);
        }
    }

    /**
     * 
     * @return
     * @throws RoutingException
     */
    protected GlobalRouterModule getRouterModule() throws RoutingException {
        if (this.routerModule == null) {
            this.log.debug("Creating router module...");
            this.routerModule = this.createRouterModule();
        }
        return this.routerModule;
    }

    /**
     * Create the Router
     * 
     * @return
     * @throws RoutingException
     */
    private GlobalRouterModule createRouterModule() throws RoutingException {
        
        GlobalRouterModule router = new GlobalRouterModule();
        router.init(this);
        
        try {
            router.init((Map<String, String>)null);
        } catch (LifeCycleException e) {
            log.warning("Initialization error", e);
            throw new RoutingException(e);
        }
        
        return router;
    }

    /**
     * Get the logger
     * 
     * @return the logger
     */
    public Logger getLogger() {
        return this.logger;
    }


    /* (non-Javadoc)
     * @see org.ow2.petals.jbi.messaging.routing.Router#addComponent(org.ow2.petals.jbi.component.context.ComponentContextImpl)
     */
    public void addComponent(ComponentContextImpl componentContext) throws RoutingException {
        this.routerModule.addComponent(componentContext);
    }


    /* (non-Javadoc)
     * @see org.ow2.petals.jbi.messaging.routing.Router#receive(org.ow2.petals.jbi.component.context.ComponentContextImpl, long)
     */
    public MessageExchangeImpl receive(ComponentContextImpl source, long timeoutMS)
            throws RoutingException {
        return this.routerModule.receive(source, timeoutMS);
    }


    /* (non-Javadoc)
     * @see org.ow2.petals.jbi.messaging.routing.Router#removeComponent(org.ow2.petals.jbi.component.context.ComponentContextImpl)
     */
    public void removeComponent(ComponentContextImpl componentContext) throws RoutingException {
        this.routerModule.removeComponent(componentContext);
    }


    /* (non-Javadoc)
     * @see org.ow2.petals.jbi.messaging.routing.Router#send(org.ow2.petals.jbi.component.context.ComponentContextImpl, org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, long)
     */
    public void send(ComponentContextImpl source, MessageExchangeImpl exchange, long timeOut)
            throws RoutingException {
        this.routerModule.send(source, exchange, timeOut);
    }


    /* (non-Javadoc)
     * @see org.ow2.petals.jbi.messaging.routing.Router#sendSync(org.ow2.petals.jbi.component.context.ComponentContextImpl, org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl, long)
     */
    public MessageExchangeImpl sendSync(ComponentContextImpl source, MessageExchangeImpl exchange,
            long timeOut) throws RoutingException {
        return this.routerModule.sendSync(source, exchange, timeOut);
    }


    /**
     * @return the log
     */
    public LoggingUtil getLog() {
        return this.log;
    }


    /**
     * @return the endpointService
     */
    public EndpointRegistry getEndpointService() {
        return this.endpointService;
    }


    /**
     * @return the topologyService
     */
    public TopologyService getTopologyService() {
        return this.topologyService;
    }


    /**
     * @return the transporter
     */
    public Transporter getTransporter() {
        return this.transporter;
    }


    /**
     * @return the configurationService
     */
    public ConfigurationService getConfigurationService() {
        return this.configurationService;
    }
}
