/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.kernel.configuration;

import java.io.Serializable;
import java.net.URI;
import java.util.Map;

/**
 * This class holds the global topology information.<br>
 * The configuration service retrieves the configuration when it is started and
 * creates an instance of this class. The instance is exposed on the distributed
 * JNDI server.
 * 
 * @author Roland NAUDIN - EbmWebsourcing
 * @since Petals 2.0
 * 
 */
public class DomainConfiguration implements Serializable {

    /**
     * 
     */
    private static final long serialVersionUID = -4951508542017620408L;

    /**
     * Domain mode enumeration.
     */
    public static enum DomainMode {
        /**
         * Stand-alone mode
         */
        STANDALONE,
        /**
         * Static distributed mode
         */
        STATIC,
        /**
         * Dynamic distributed mode
         */
        DYNAMIC
    };

    /**
     * Domain name.
     */
    private String name;

    /**
     * Domain description
     */
    private String description;

    /**
     * The domain mode.
     */
    private DomainMode mode;

    // ************ Network *************//

    /**
     * Network multicast IP. IP used to transmit PEtALS multicast messages
     */
    private String networkMuticastIp;

    /**
     * Network multicast port. Port used to transmit PEtALS multicast messages
     */
    private int networkMuticastPort;

    /**
     * Network unicast address map. Map of id/unicast addresses of the members
     * of the PEtALS network.
     */
    private Map<String, String> networkUnicastAddressMap;

    // ************* JNDI **************//

    /**
     * JNDI Factory
     */
    private String jndiFactory;

    /**
     * JNDI URL
     */
    private URI jndiProviderUrl;

    /***
     * JNDI security Principal name
     */
    private String jndiSecurityPrincipal;

    /**
     * JNDI security Credentials
     */
    private String jndiSecurityCredentials;

    /**
     * Connection Pool size of the JNDI server
     */
    private int jndiPoolSize;

    /**
     * Batch size of the JNDI server
     */
    private int jndiBatchSize;

    public String getName() {
        return this.name;
    }

    public String getDescription() {
        return this.description;
    }

    public DomainMode getMode() {
        return this.mode;
    }

    public String getNetworkMuticastIp() {
        return this.networkMuticastIp;
    }

    public int getNetworkMuticastPort() {
        return this.networkMuticastPort;
    }

    public Map<String, String> getNetworkUnicastAddressMap() {
        return this.networkUnicastAddressMap;
    }

    public String getJndiFactory() {
        return this.jndiFactory;
    }

    @Override
    public String toString() {
        StringBuilder toStringBuilder = new StringBuilder();
        toStringBuilder.append("Domain " + this.name + "\n\t\t\t  Configuration :");
        toStringBuilder.append("\n\t\t\t  - Mode:" + this.mode);
        if (this.networkMuticastIp != null) {
            toStringBuilder.append("\n\t\t\t  - Network:multicast UDP");
            toStringBuilder.append("\n\t\t\t  - MulticastIP:" + this.networkMuticastIp);
            toStringBuilder.append("\n\t\t\t  - MulticastPort:" + this.networkMuticastPort);
        } else {
            toStringBuilder.append("\n\t\t\t  - Network:unicast TCP");
        }

        if (this.jndiFactory == null) {
            toStringBuilder.append("\n\t\t\t  - JNDIFactory: None");
        } else {
            toStringBuilder.append("\n\t\t\t  - JNDIFactory:" + this.jndiFactory);
        }

        if (this.jndiProviderUrl != null) {
            toStringBuilder.append("\n\t\t\t  - JNDIProviderUrl:" + this.jndiProviderUrl);
        }

        if (this.jndiSecurityPrincipal != null) {
            toStringBuilder.append("\n\t\t\t  - JNDISecurityPrincipal:"
                    + this.jndiSecurityPrincipal);
        }

        if (this.jndiSecurityCredentials != null) {
            toStringBuilder.append("\n\t\t\t  - JNDISecurityCredentials:"
                    + this.jndiSecurityCredentials);
        }

        if (this.jndiPoolSize != 0) {
            toStringBuilder.append("\n\t\t\t  - JNDIPool size:" + this.jndiPoolSize);
        }

        if (this.jndiBatchSize != 0) {
            toStringBuilder.append("\n\t\t\t  - JNDIbatch size:" + this.jndiBatchSize);
        }

        return toStringBuilder.toString();
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((this.name == null) ? 0 : this.name.hashCode());
        result = prime * result + ((this.description == null) ? 0 : this.description.hashCode());
        result = prime * result + ((this.jndiFactory == null) ? 0 : this.jndiFactory.hashCode());
        result = prime * result
                + ((this.jndiProviderUrl == null) ? 0 : this.jndiProviderUrl.hashCode());
        result = prime
                * result
                + ((this.jndiSecurityPrincipal == null) ? 0 : this.jndiSecurityPrincipal.hashCode());
        result = prime
                * result
                + ((this.jndiSecurityCredentials == null) ? 0 : this.jndiSecurityCredentials
                        .hashCode());
        result = prime * result + this.jndiPoolSize + this.jndiBatchSize;
        result = prime * result + ((this.mode == null) ? 0 : this.mode.hashCode());
        result = prime * result
                + ((this.networkMuticastIp == null) ? 0 : this.networkMuticastIp.hashCode());
        result = prime * result + this.networkMuticastPort;
        result = prime
                * result
                + ((this.networkUnicastAddressMap == null) ? 0 : this.networkUnicastAddressMap
                        .hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }

        if (obj instanceof DomainConfiguration) {
            final DomainConfiguration otherDomainConfiguration = (DomainConfiguration) obj;

            // mandatory fields
            if (!(this.name.equals(otherDomainConfiguration.name) && this.mode
                    .equals(otherDomainConfiguration.mode))) {
                return false;
            }

            // optional fields
            if (this.description == null && otherDomainConfiguration.description != null) {
                return false;
            } else if (!this.description.equals(otherDomainConfiguration.description)) {
                return false;
            }
            if (this.jndiFactory == null && otherDomainConfiguration.jndiFactory != null) {
                return false;
            } else if (!(this.jndiFactory.equals(otherDomainConfiguration.jndiFactory) && (this.jndiProviderUrl
                    .equals(otherDomainConfiguration.jndiProviderUrl) && this.jndiSecurityPrincipal
                    .equals(otherDomainConfiguration.jndiSecurityPrincipal)))) {
                return false;
            }

            if (this.networkMuticastIp == null
                    && otherDomainConfiguration.networkMuticastIp != null) {
                return false;
            } else if (!(this.networkMuticastIp.equals(otherDomainConfiguration.networkMuticastIp) && this.networkMuticastPort == otherDomainConfiguration.networkMuticastPort)) {
                return false;
            }

            if (this.networkUnicastAddressMap == null
                    && otherDomainConfiguration.networkUnicastAddressMap != null) {
                return false;
            } else if (!this.networkUnicastAddressMap
                    .equals(otherDomainConfiguration.networkUnicastAddressMap)) {
                return false;
            }
        } else {
            return false;
        }

        return true;
    }

    public URI getJndiProviderUrl() {
        return jndiProviderUrl;
    }

    public String getJndiSecurityPrincipal() {
        return jndiSecurityPrincipal;
    }

    public String getJndiSecurityCredentials() {
        return jndiSecurityCredentials;
    }

    public int getJndiPoolSize() {
        return jndiPoolSize;
    }

    public int getJndiBatchSize() {
        return jndiBatchSize;
    }

    protected void setName(String name) {
        this.name = name;
    }

    protected void setDescription(String description) {
        this.description = description;
    }

    protected void setMode(DomainMode mode) {
        this.mode = mode;
    }

    protected void setNetworkMuticastIp(String networkMuticastIp) {
        this.networkMuticastIp = networkMuticastIp;
    }

    protected void setNetworkMuticastPort(int networkMuticastPort) {
        this.networkMuticastPort = networkMuticastPort;
    }

    protected void setNetworkUnicastAddressMap(Map<String, String> networkUnicastAddressMap) {
        this.networkUnicastAddressMap = networkUnicastAddressMap;
    }

    protected void setJndiFactory(String jndiFactory) {
        this.jndiFactory = jndiFactory;
    }

    protected void setJndiProviderUrl(URI jndiProviderUrl) {
        this.jndiProviderUrl = jndiProviderUrl;
    }

    protected void setJndiSecurityPrincipal(String jndiSecurityPrincipal) {
        this.jndiSecurityPrincipal = jndiSecurityPrincipal;
    }

    protected void setJndiSecurityCredentials(String jndiSecurityCredentials) {
        this.jndiSecurityCredentials = jndiSecurityCredentials;
    }

    protected void setJndiPoolSize(int poolSize) {
        this.jndiPoolSize = poolSize;
    }

    protected void setJndiBatchSize(int batchSize) {
        this.jndiBatchSize = batchSize;
    }

}
