/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 eBM Websourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: RepositoryServiceImpl.java 11031 2009-05-12 08:04:05Z rnaudin $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.platform.repository;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URI;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.filefilter.PrefixFileFilter;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.api.server.util.SystemUtil;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.ParameterCheckHelper;
import org.ow2.petals.util.ZipUtil;

/**
 * Platform Component Repository service implementation.
 * 
 * @version $Rev: 11031 $ $Date: 2009-05-12 10:04:05 +0200 (Tue, 12 May 2009) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 * @author alouis - EBM Websourcing
 * @author ofabre - EBM Websourcing
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.platform.repository.RepositoryService.class))
public class RepositoryServiceImpl implements RepositoryService {

    /**
     * The date pattern
     */
    private static final String DATE_PATTERN = "MM-dd-yyyy-hhmmssSSS";

    private static final int DATE_PATTERN_LENGTH = 20;

    /**
     * Path to component install directory
     */
    private static final String INSTALL_PATH = "install";

    /**
     * Path to components directory into petals repository
     */
    private static final String COMPONENTS_PATH = "components";

    /**
     * Path to components directory into petals repository
     */
    private static final String SHAREDLIBS_PATH = "shared-libraries";

    /**
     * Path to lost+found directory into petals repository
     */
    private static final String LOSTPLUSFOUND_PATH = "lost+found";

    /**
     * Path to components directory into petals repository
     */
    private static final String SERVICEASSEMBLIES_PATH = "service-assemblies";

    /**
     * Path to component working directory
     */
    private static final String WORK_PATH = "work";

    /**
     * The complete repository directory path.
     */
    private File repositoryDirectory = null;

    /**
     * Entity types (components, shared libs or service assemblies)
     */
    public static enum EntityType {
        COMPONENT_TYPE, SL_TYPE, SA_TYPE
    };

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * DateFormat used for repository folders
     */
    private SimpleDateFormat folderDateFormat;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    /**
     * Copy an entity package into petals repository
     * 
     * @param entityId
     *            {@link String} unique entity id that identifies an unique
     *            entity directory into the petals repository.
     * @param tempFile
     *            {@link File} the entity package file to add into the
     *            repository.
     * @param entityType
     *            the entity type: component, shared lib or service assembly
     * @return installation root {@link URI}
     * @throws PetalsException
     *             if entity installation or work root can't be created or if
     *             entity files can't be copied in the installation root
     */
    private URI addPackage(String entityId, File tempFile, EntityType entityType)
            throws PetalsException {

        File rootFile = this.createEntityRootFile(entityId, entityType);

        File installationRoot = this.createEntityInstallRootFolder(rootFile);

        File workRoot = null;
        try {
            workRoot = this.createEntityWorkRootFolder(rootFile);
        } catch (PetalsException e) {
            try {
                FileUtils.forceDelete(installationRoot);
            } catch (IOException e1) {
                // Nothing to do, just log
                this.log.warning("Residual incomplete component folder");
            }
            throw e;
        }

        try {
            FileUtils.copyDirectory(tempFile, installationRoot);
            FileUtils.forceDelete(tempFile);
        } catch (IOException e) {
            try {
                FileUtils.forceDelete(installationRoot);
                FileUtils.forceDelete(workRoot);
            } catch (IOException e1) {
                // Nothing to do, just log
                this.log.warning("Residual incomplete component folder");
            }
            throw new PetalsException(e);
        }

        return installationRoot.toURI();
    }

    private File createEntityWorkRootFolder(File rootFile) throws PetalsException {
        File workRoot = new File(rootFile, WORK_PATH);
        if (!workRoot.mkdirs()) {
            throw new PetalsException("Failed to create work dir: " + workRoot.getAbsolutePath());
        }
        return workRoot;
    }

    private File createEntityInstallRootFolder(File rootFile) throws PetalsException {
        File installationRoot = new File(rootFile, INSTALL_PATH);

        if (!installationRoot.mkdirs()) {
            throw new PetalsException("Failed to create installation root dir: "
                    + installationRoot.getAbsolutePath());
        }
        return installationRoot;
    }

    private File createEntityRootFile(String entityId, EntityType entityType)
            throws PetalsException {
        File rootFile = null;

        switch (entityType) {
            case COMPONENT_TYPE:
                rootFile = this.createNewComponentRoot(entityId);
                break;
            case SA_TYPE:
                rootFile = this.createNewServiceAssemblyRoot(entityId);
                break;
            case SL_TYPE:
                rootFile = this.createNewSharedLibRoot(entityId);
                break;
        }
        return rootFile;
    }

    public URI addComponentPackage(String componentID, File tempFile) throws PetalsException {
        ParameterCheckHelper.isNullOrEmptyParameter(componentID, "componentID");
        ParameterCheckHelper.isNullParameter(tempFile, "tempFile");
        return this.addPackage(componentID, tempFile, EntityType.COMPONENT_TYPE);
    }

    public URI addSharedLibPackage(String sharedLibID, File tempFile) throws PetalsException {
        ParameterCheckHelper.isNullOrEmptyParameter(sharedLibID, "sharedLibID");
        ParameterCheckHelper.isNullParameter(tempFile, "tempFile");
        return this.addPackage(sharedLibID, tempFile, EntityType.SL_TYPE);
    }

    public URI addServiceAssemblyPackage(String serviceAssemblyID, File tempFile)
            throws PetalsException {
        ParameterCheckHelper.isNullOrEmptyParameter(serviceAssemblyID, "serviceAssemblyID");
        ParameterCheckHelper.isNullParameter(tempFile, "tempFile");
        return this.addPackage(serviceAssemblyID, tempFile, EntityType.SA_TYPE);
    }

    /**
     * @see org.ow2.petals.platform.repository.RepositoryService#
     *      explodeSUIntoSAInstallFolder(String, URI, String)
     */
    public URI explodeSUIntoSAInstallFolder(String suId, URI suZipLocation, String saId)
            throws PetalsException {
        File installationRoot = this.createSUInstallRootFolder(suId, saId);
        try {
            File zipArchive = new File(suZipLocation);
            ZipUtil.unzipArchive(zipArchive, installationRoot);
        } catch (IOException e) {
            try {
                FileUtils.forceDelete(installationRoot);
            } catch (IOException e1) {
                // Nothing to do, just log
                this.log.warning("Residual incomplete component folder");
            }
            throw new PetalsException(e);
        }
        return installationRoot.toURI();
    }

    /**
     * Create the SU install root into the SA install root
     * 
     * @param suId
     *            the service unit unique identifier
     * @param saId
     *            the service assembly unique identifier
     * @return the SU install root File
     * @throws PetalsException
     *             if SU install root can't be created
     */
    private File createSUInstallRootFolder(String suId, String saId) throws PetalsException {
        File installationRoot = new File(this.getServiceAssemblyInstallRoot(saId), suId);

        if (!installationRoot.exists() && !installationRoot.mkdirs()) {
            throw new PetalsException("Failed to create installation root dir: "
                    + installationRoot.getAbsolutePath());
        }
        return installationRoot;
    }

    public File getComponentsDirectory() {
        return new File(this.getRepositoryDirectory(), File.separator + COMPONENTS_PATH);

    }

    public File getServiceAssembliesDirectory() {
        return new File(this.getRepositoryDirectory(), File.separator + SERVICEASSEMBLIES_PATH);

    }

    public File getSharedLibsDirectory() {
        return new File(this.getRepositoryDirectory(), File.separator + SHAREDLIBS_PATH);
    }

    public String getLostPlusFoundDirectory() {
        return SystemUtil.getPetalsInstallDirectory().getAbsolutePath() + File.separator
                + LOSTPLUSFOUND_PATH;
    }

    public File getComponentInstallRoot(String componentId) throws PetalsException {
        return new File(this.getComponentRoot(componentId), INSTALL_PATH);

    }

    public File getComponentRoot(String componentId) throws PetalsException {
        return this.getEntityRoot(componentId, EntityType.COMPONENT_TYPE);
    }

    private String[] filterPathsOnlength(String componentId, String[] componentsFilePaths) {
        List<String> goodFilePaths = new ArrayList<String>();

        for (String path : componentsFilePaths) {
            if (path.length() == componentId.length() + DATE_PATTERN_LENGTH + 1) {
                goodFilePaths.add(path);
            }
        }

        return goodFilePaths.toArray(new String[0]);
    }

    /**
     * Retrieve the most recent path in an array of folder path, according to
     * the included date information.
     * 
     * @param filePaths
     *            the file paths to compare
     * @param entityId
     *            an entity (component, shared lib or service assembly) id
     * @return the most recent file path, null if given filePaths array is null
     * @throws ParseException
     *             if an error occurs during Date extraction
     */
    private String searchForMostRecentFilePath(String[] filePaths, String entityId)
            throws ParseException {
        String mostRecentPath = null;
        Date mostRecentDate = null;
        if (filePaths != null) {
            for (String path : filePaths) {
                String baseName = FilenameUtils.getName(path);
                Date currentDate = this.extractDateFromName(baseName, entityId);
                if (mostRecentDate != null) {
                    if (currentDate.after(mostRecentDate)) {
                        mostRecentDate = currentDate;
                        mostRecentPath = path;
                    }
                } else {
                    mostRecentDate = currentDate;
                    mostRecentPath = path;
                }
            }
        }
        return mostRecentPath;
    }

    /**
     * Extract a Date from a file name.
     * 
     * @param baseName
     *            a file name
     * @param entityId
     *            an entity (component, shared lib or service assembly) id
     * @return the file Date, non null
     * @throws ParseException
     *             if an error occurs during Date extraction
     */
    private Date extractDateFromName(String baseName, String entityId) throws ParseException {
        String dateString = baseName.substring(baseName.indexOf(entityId) + entityId.length() + 1);
        Date result;
        synchronized (this.folderDateFormat) {
            result = this.folderDateFormat.parse(dateString);
        }

        return result;
    }

    private File createNewComponentRoot(String componentId) throws PetalsException {
        String name = componentId + "-"
                + this.folderDateFormat.format(new Date(System.currentTimeMillis()));

        return new File(this.getComponentsDirectory(), name);
    }

    public File getComponentWorkRoot(String componentId) throws PetalsException {
        return new File(this.getComponentRoot(componentId), WORK_PATH);

    }

    public File getSharedLibRoot(String slId) throws PetalsException {
        return this.getEntityRoot(slId, EntityType.SL_TYPE);
    }

    private File createNewSharedLibRoot(String slId) throws PetalsException {
        String name = slId + "-"
                + this.folderDateFormat.format(new Date(System.currentTimeMillis()));

        return new File(this.getSharedLibsDirectory(), name);
    }

    public File getSharedLibWorkRoot(String slId) throws PetalsException {
        return new File(this.getSharedLibRoot(slId), WORK_PATH);

    }

    public File getSharedLibInstallRoot(String slId) throws PetalsException {
        String name = INSTALL_PATH;
        return new File(this.getSharedLibRoot(slId), name);
    }

    private File getEntityRoot(String entityId, EntityType entityType) throws PetalsException {
        File result = null;
        File entitiesDirectory = null;

        switch (entityType) {
            case COMPONENT_TYPE:
                entitiesDirectory = this.getComponentsDirectory();
                break;
            case SA_TYPE:
                entitiesDirectory = this.getServiceAssembliesDirectory();
                break;
            case SL_TYPE:
                entitiesDirectory = this.getSharedLibsDirectory();
                break;
        }

        String[] entityFilePaths = entitiesDirectory.list(new PrefixFileFilter(entityId));

        String[] entityFilePathsWithGoodLength = this
                .filterPathsOnlength(entityId, entityFilePaths);

        String selectedPath = null;
        try {
            selectedPath = this
                    .searchForMostRecentFilePath(entityFilePathsWithGoodLength, entityId);
        } catch (ParseException e) {
            throw new PetalsException(
                    "Can't retrieve entity root folder for the following entity: " + entityId, e);
        }
        if (selectedPath != null) {
            result = new File(entitiesDirectory, selectedPath);
        }
        return result;
    }

    public File getServiceAssemblyRoot(String saId) throws PetalsException {
        return this.getEntityRoot(saId, EntityType.SA_TYPE);
    }

    private File createNewServiceAssemblyRoot(String saId) throws PetalsException {
        String name = saId + "-"
                + this.folderDateFormat.format(new Date(System.currentTimeMillis()));

        return new File(this.getServiceAssembliesDirectory(), name);
    }

    public File getServiceAssemblyWorkRoot(String saId) throws PetalsException {
        return new File(this.getServiceAssemblyRoot(saId), WORK_PATH);
    }

    public File getServiceAssemblyInstallRoot(String saId) throws PetalsException {
        return new File(this.getServiceAssemblyRoot(saId), INSTALL_PATH);
    }

    /**
     * @throws PetalsException
     * @see org.ow2.petals.platform.repository.RepositoryService#removePackage(java
     *      .lang.String)
     */
    private boolean removePackage(String entityID, EntityType entityType) throws PetalsException {
        boolean removed = true;
        File entityRoot = null;
        switch (entityType) {
            case COMPONENT_TYPE:
                entityRoot = this.getComponentRoot(entityID);
                break;
            case SA_TYPE:
                entityRoot = this.getServiceAssemblyRoot(entityID);
                break;
            case SL_TYPE:
                entityRoot = this.getSharedLibRoot(entityID);
                break;
        }
        try {
            if (entityRoot != null) {
                FileUtils.deleteDirectory(entityRoot);
            }
        } catch (IOException e) {
            this.log.warning("Failed to delete old Petals entity folder. "
                    + "This folder will be deleted on a future Petals restarting.");
        }
        return removed;
    }

    public boolean removeComponentPackage(String componentId) throws PetalsException {
        return this.removePackage(componentId, EntityType.COMPONENT_TYPE);
    }

    public boolean removeSharedLibPackage(String slId) throws PetalsException {
        return this.removePackage(slId, EntityType.SL_TYPE);
    }

    public boolean removeServiceAssemblyPackage(String saId) throws PetalsException {
        return this.removePackage(saId, EntityType.SA_TYPE);
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws FileNotFoundException {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        this.folderDateFormat = new SimpleDateFormat(DATE_PATTERN);
        this.containerConfiguration = this.configurationService.getContainerConfiguration();

        // Initialization
        /*
         * Check if a repository folder path is specified in the
         * 'server.properties' file. If it is, create it as specified
         */
        this.repositoryDirectory = new File(this.containerConfiguration
                .getRepositoryDirectoryPath());

        try {
            if (!this.repositoryDirectory.exists()) {
                throw new FileNotFoundException("The repository directory does not exist:"
                        + this.containerConfiguration.getRepositoryDirectoryPath());
            }

            if (!this.getComponentsDirectory().exists())
                throw new FileNotFoundException(
                        "The component repository of the server.properties "
                                + "file must be a valid file: " + this.getComponentsDirectory());

            if (!this.getServiceAssembliesDirectory().exists())
                throw new FileNotFoundException(
                        "The service assemblies repository of the server.properties "
                                + "file must be a valid file: "
                                + this.getServiceAssembliesDirectory());

            if (!this.getSharedLibsDirectory().exists())
                throw new FileNotFoundException(
                        "The shared libraries repository of the server.properties "
                                + "file must be a valid file: " + this.getSharedLibsDirectory());
        } catch (FileNotFoundException e) {
            this.log.error(e);
            throw e;
        }

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.call();
    }

    /**
     * Return the petals repository directory. It will be created when this
     * method is called for the first time
     * 
     * @throws FileNotFoundException
     * 
     * @throws RuntimeException
     *             if the specified or default repository can't be created
     */
    public File getRepositoryDirectory() {
        return this.repositoryDirectory;
    }

}
