/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SystemStateServiceImpl.java 154 27 avr. 2006 ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.platform.systemstate;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.platform.repository.RepositoryService;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Creates and reads state files for components, shared libs and service
 * assemblies. State files are text files that keep the state of JBI entities to
 * allow state retreval after Petals system shutdown or crash.
 * 
 * @author Adrien LOUIS - EBM Websourcing
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.platform.systemstate.SystemStateService.class))
public class SystemStateServiceImpl implements SystemStateService {

    protected static final String NS = "http://org.objectweb.petals/xml/ns/system-state";

    protected static final String SYSTEMSTATE_XSD = "system-state.xsd";

    protected static final String SYSTEMSTATE_FILE = "system-state.xml";

    protected Document document;

    protected File systemStateFile = null;

    @Requires(name = "repository", signature = org.ow2.petals.platform.repository.RepositoryService.class)
    protected RepositoryService repositorySrv;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name = "logger")
    protected Logger logger;

    public SystemStateServiceImpl() {
    }

    public ComponentState createComponentStateHolder(String componentName, String installURL,
            String zipURL, String installState, String lifecycleState) throws Exception {
        Node e = this.createComponentNode(componentName, installURL, zipURL, installState,
                lifecycleState);

        this.getComponentsNode().appendChild(e);

        this.saveSystemState();

        return this.createComponent(e);
    }

    // ///////////////////////////////////////////////////////////////////////
    //
    // Document <-> File management methods
    //
    // ///////////////////////////////////////////////////////////////////////

    public ServiceAssemblyState createServiceAssemblyStateHolder(String saName, String installURL,
            String zipURL, String lifecycleState) throws Exception {
        Node e = this.createServiceAssemblyNode(saName, installURL, zipURL, lifecycleState);

        this.getServiceAssembliesNode().appendChild(e);

        this.saveSystemState();

        return this.createServiceAssembly(e);
    }

    public SharedLibraryState createSharedLibraryStateHolder(String slName, String installURL,
            String zipURL) throws Exception {
        Node e = this.createSharedLibraryNode(slName, installURL, zipURL);

        this.getSharedLibrariesNode().appendChild(e);

        this.saveSystemState();

        return this.createSharedLibrary(e);
    }

    public ComponentState deleteComponentStateHolder(String componentName) throws Exception {
        Node n = this.getComponentNode(componentName);

        this.deleteStateAndSave(this.getComponentsNode(), n);

        return this.createComponent(n);
    }

    public ServiceAssemblyState deleteServiceAssemblyStateHolder(String saName) throws Exception {
        Node n = this.getServiceAssemblyNode(saName);

        this.deleteStateAndSave(this.getServiceAssembliesNode(), n);

        return this.createServiceAssembly(n);
    }

    // ///////////////////////////////////////////////////////////////////////
    //
    // getNodeXXX methods
    //
    // ///////////////////////////////////////////////////////////////////////

    public SharedLibraryState deleteSharedLibraryStateHolder(String slName) throws Exception {
        Node n = this.getSharedLibraryNode(slName);

        this.deleteStateAndSave(this.getSharedLibrariesNode(), n);

        return this.createSharedLibrary(n);
    }

    /**
     * Initialize the SystemStateServiceImpl object. The recovery-file is
     * created or read, then this object is setup with the information contained
     * in the file.
     * 
     * @throws IOException
     *             problems accessing the file
     * @throws SAXException
     *             the structure of the xml-file is not valid.
     */
    public void init() {
        try {
            this.loadStateStorageDocument();
        } catch (Exception e) {
            this.log.error("Can't load system-state file", e);
        }
        try {
            this.validateSystemStateXml();
        } catch (SAXException e) {
            this.log.error("Can't validate system-state file", e);
        } catch (IOException e) {
            this.log.error("Can't validate system-state file", e);
        }
    }

    public List<ComponentState> retrieveAllComponentStates() {
        NodeList elements = this.getComponentsNode().getChildNodes();

        List<ComponentState> list = new ArrayList<ComponentState>();

        for (int i = 0; i < elements.getLength(); i++) {
            list.add(this.createComponent(elements.item(i)));
        }
        return list;
    }

    public List<ServiceAssemblyState> retrieveAllServiceAssemblyStates() {
        NodeList elements = this.getServiceAssembliesNode().getChildNodes();

        List<ServiceAssemblyState> list = new ArrayList<ServiceAssemblyState>();

        for (int i = 0; i < elements.getLength(); i++) {
            list.add(this.createServiceAssembly(elements.item(i)));
        }
        return list;
    }

    public List<SharedLibraryState> retrieveAllSharedLibraryStates() {
        NodeList elements = this.getSharedLibrariesNode().getChildNodes();

        List<SharedLibraryState> list = new ArrayList<SharedLibraryState>();

        for (int i = 0; i < elements.getLength(); i++) {
            list.add(this.createSharedLibrary(elements.item(i)));
        }
        return list;
    }

    public void updateComponentInstallationState(String componentName, String installState)
            throws Exception {
        Node n = this.getComponentNode(componentName);

        this.updateStateAndSave(n, INSTALL_STATE_ATTRIBUTE, installState);
    }

    public void updateComponentLifeCycleState(String componentName, String lifecycleState)
            throws Exception {
        Node n = this.getComponentNode(componentName);

        this.updateStateAndSave(n, LIFECYCLE_STATE_ATTRIBUTE, lifecycleState);
    }

    public void updateServiceAssemblyState(String saName, String lifecycleState) throws Exception {
        Node n = this.getServiceAssemblyNode(saName);

        this.updateStateAndSave(n, LIFECYCLE_STATE_ATTRIBUTE, lifecycleState);
    }

    /**
     * 
     * @param sourceURI
     * @return
     * @throws SAXException
     * @throws IOException
     */
    public void validateSystemStateXml() throws SAXException, IOException {
        // create a SchemaFactory capable of understanding WXS schemas
        SchemaFactory factory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

        // load a WXS schema, represented by a Schema instance
        Source schemaSource = new StreamSource(this.getClass().getResourceAsStream(
                "/" + SYSTEMSTATE_XSD));
        Schema schema = factory.newSchema(schemaSource);

        // create a Validator instance which is be used to validate a document
        Validator validator = schema.newValidator();

        // validate
        StreamSource streamSource = new StreamSource(systemStateFile);
        validator.validate(streamSource);
        if (streamSource.getInputStream() != null) {
            streamSource.getInputStream().close();
        }
    }

    public ComponentState getComponentState(String componentName) {
        return this.createComponent(this.getComponentNode(componentName));
    }

    public ServiceAssemblyState getServiceAssemblyState(String saName) {
        return this.createServiceAssembly(this.getServiceAssemblyNode(saName));
    }

    public SharedLibraryState getSharedLibraryState(String slName) {
        return this.createSharedLibrary(this.getSharedLibraryNode(slName));
    }

    public List<String> getAllDeployedServiceAssemblyNames() {
        List<String> names = new ArrayList<String>();
        for (ServiceAssemblyState state : this.retrieveAllServiceAssemblyStates()) {
            names.add(state.getName());
        }
        return names;
    }

    public List<String> getAllLoadedComponentNames() {
        List<String> names = new ArrayList<String>();
        for (ComponentState state : this.retrieveAllComponentStates()) {
            names.add(state.getName());
        }
        return names;
    }

    public List<String> getAllInstalledSharedLibNames() {
        List<String> names = new ArrayList<String>();
        for (SharedLibraryState state : this.retrieveAllSharedLibraryStates()) {
            names.add(state.getName());
        }
        return names;
    }

    public boolean isComponentLoaded(String componentName) {
        return this.getAllLoadedComponentNames().contains(componentName);
    }

    public boolean isSLLoaded(String slName) {
        return this.getAllInstalledSharedLibNames().contains(slName);
    }

    public boolean isSADeployed(String saName) {
        return this.getAllDeployedServiceAssemblyNames().contains(saName);
    }

    /**
     * Create a ComponentState from the given node
     * 
     * @param node
     * @return a ComponentState , null if node is null
     */
    protected ComponentState createComponent(Node node) {
        String name = XMLUtil.getAttributeValue(node, NAME_ATTRIBUTE);
        String iURL = XMLUtil.getAttributeValue(node, INSTALL_PATH_ATTRIBUTE);
        String aURL = XMLUtil.getAttributeValue(node, ARCHIVE_PATH_ATTRIBUTE);
        String iSte = XMLUtil.getAttributeValue(node, INSTALL_STATE_ATTRIBUTE);
        String lSte = XMLUtil.getAttributeValue(node, LIFECYCLE_STATE_ATTRIBUTE);

        return new ComponentState(name, iURL, aURL, iSte, lSte);
    }

    /**
     * create a Node representing a component state
     * 
     * @param componentName
     * @param installURL
     * @param zipURL
     * @param installState
     * @param lifecycleState
     * @return
     * @throws Exception
     */
    protected Node createComponentNode(String componentName, String installURL, String zipURL,
            String installState, String lifecycleState) throws Exception {
        Node element = XMLUtil.createNode(this.document, COMPONENT_ELEMENT, NAME_ATTRIBUTE,
                componentName, INSTALL_PATH_ATTRIBUTE, installURL, ARCHIVE_PATH_ATTRIBUTE, zipURL,
                INSTALL_STATE_ATTRIBUTE, installState, LIFECYCLE_STATE_ATTRIBUTE, lifecycleState);

        return element;
    }

    /**
     * Create a ServiceAssemblyState from the given node
     * 
     * @param node
     * @return a ServiceAssemblyState , null if node is null
     */
    protected ServiceAssemblyState createServiceAssembly(Node node) {
        String name = XMLUtil.getAttributeValue(node, NAME_ATTRIBUTE);
        String iURL = XMLUtil.getAttributeValue(node, INSTALL_PATH_ATTRIBUTE);
        String aURL = XMLUtil.getAttributeValue(node, ARCHIVE_PATH_ATTRIBUTE);
        String lSte = XMLUtil.getAttributeValue(node, LIFECYCLE_STATE_ATTRIBUTE);

        return new ServiceAssemblyState(name, iURL, aURL, lSte);
    }

    /**
     * create a Node representing a sa state
     * 
     * @param saName
     * @param installURL
     * @param zipURL
     * @param lifecycleState
     * @return
     * @throws Exception
     */
    protected Node createServiceAssemblyNode(String saName, String installURL, String zipURL,
            String lifecycleState) throws Exception {
        Node element = XMLUtil.createNode(this.document, SERVICE_ASSEMBLY_ELEMENT, NAME_ATTRIBUTE,
                saName, INSTALL_PATH_ATTRIBUTE, installURL, ARCHIVE_PATH_ATTRIBUTE, zipURL,
                LIFECYCLE_STATE_ATTRIBUTE, lifecycleState);

        return element;
    }

    /**
     * Create a SharedLibraryState from the given node
     * 
     * @param node
     * @return a SharedLibraryState , null if node is null
     */
    protected SharedLibraryState createSharedLibrary(Node node) {
        String name = XMLUtil.getAttributeValue(node, NAME_ATTRIBUTE);
        String iURL = XMLUtil.getAttributeValue(node, INSTALL_PATH_ATTRIBUTE);
        String aURL = XMLUtil.getAttributeValue(node, ARCHIVE_PATH_ATTRIBUTE);

        return new SharedLibraryState(name, iURL, aURL);
    }

    /**
     * create a Node representing a sl state
     * 
     * @param slName
     * @param installURL
     * @param zipURL
     * @return
     * @throws Exception
     */
    protected Node createSharedLibraryNode(String slName, String installURL, String zipURL)
            throws Exception {

        Node element = XMLUtil.createNode(this.document, SHARE_LIBRARY_ELEMENT, NAME_ATTRIBUTE, slName,
                INSTALL_PATH_ATTRIBUTE, installURL, ARCHIVE_PATH_ATTRIBUTE, zipURL);

        return element;
    }

    // ///////////////////////////////////////////////////////////////////////
    //
    // service methods
    //
    // ///////////////////////////////////////////////////////////////////////

    /**
     * delete the element, and save the Document
     * 
     * @param parent
     * @param child
     * @throws Exception
     *             node not found, saving problem
     */
    protected void deleteStateAndSave(Node parent, Node child) throws Exception {
        if (parent != null && child != null) {
            parent.removeChild(child);
            saveSystemState();
        }
    }

    protected Node getComponentNode(String name) {
        return XMLUtil.getNodeWithAttribute(this.document, COMPONENT_ELEMENT, "name", name);
    }

    protected Node getComponentsNode() {
        return XMLUtil.getNode(this.document, COMPONENTS_ELEMENT);
    }

    protected Node getServiceAssembliesNode() {
        return XMLUtil.getNode(this.document, SERVICE_ASSEMBLIES_ELEMENT);
    }

    protected Node getServiceAssemblyNode(String name) {
        return XMLUtil.getNodeWithAttribute(this.document, SERVICE_ASSEMBLY_ELEMENT, "name", name);
    }

    protected Node getSharedLibrariesNode() {
        return XMLUtil.getNode(this.document, SHARED_LIBRARIES_ELEMENT);
    }

    protected Node getSharedLibraryNode(String name) {
        return XMLUtil.getNodeWithAttribute(this.document, SHARE_LIBRARY_ELEMENT, "name", name);
    }

    /**
     * add to the empty Document the necessary elements. clean the document
     * before if necessary
     * 
     */
    protected void initializeDocument() {
        NodeList children = this.document.getChildNodes();

        for (int i = 0; i < children.getLength(); i++) {
            this.document.removeChild(children.item(i));
        }

        // document.setXmlVersion("1.0");

        Element s = this.document.createElement(SYSTEM_STATE_ELEMENT);
        Node c = this.document.createElement(COMPONENTS_ELEMENT);
        Node sl = this.document.createElement(SHARED_LIBRARIES_ELEMENT);
        Node sa = this.document.createElement(SERVICE_ASSEMBLIES_ELEMENT);

        s.setAttribute(XMLConstants.XMLNS_ATTRIBUTE, SystemStateServiceImpl.NS);
        this.document.appendChild(s);
        s.appendChild(c);
        s.appendChild(sl);
        s.appendChild(sa);
    }

    /**
     * Load the Document from the systemstate file. if the file is empty, the
     * document is initialized and saved
     */
    protected void loadStateStorageDocument() throws Exception {
        // create file if necessary
        if (!systemStateFile.exists()) {
            systemStateFile.createNewFile();
        }
        DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
        // if file is empty, initialize it (test the length of the file)
        FileInputStream fileInputStream = new FileInputStream(systemStateFile);
        if (fileInputStream.available() == 0) {
            document = docBuilder.newDocument();
            initializeDocument();
            saveSystemState();
        } else {
            this.document = docBuilder.parse(this.systemStateFile);
        }
        fileInputStream.close();
    }

    /**
     * Write the document representing the system state to the recovery-file
     * 
     * @throws Exception
     */
    protected void saveSystemState() throws Exception {
        Transformer transformer = TransformerFactory.newInstance().newTransformer();
        FileWriter fileWriter = new FileWriter(systemStateFile);
        StreamResult streamResult= new StreamResult(fileWriter);
        transformer.transform(new DOMSource(this.document), streamResult);
        if (streamResult.getOutputStream() != null) {
            streamResult.getOutputStream().close();
        }
        fileWriter.close();
    }

    /**
     * update the value of the attribute of the node, and save the Document
     * 
     * @param node
     * @param att
     * @param value
     * @throws Exception
     *             node not found, saving problem
     */
    protected void updateStateAndSave(Node node, String att, String value) throws Exception {
        node.getAttributes().getNamedItem(att).setNodeValue(value);

        this.saveSystemState();
    }

    /**
     * Initialize the SystemStateService service
     * 
     * @throws URISyntaxException
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() {
        this.log = new LoggingUtil(this.logger);
        this.log.start();
        File repoPath = this.repositorySrv.getRepositoryDirectory();

        if (repoPath.exists()) {
            repoPath.mkdirs();
        }
        systemStateFile = new File(repoPath.getAbsolutePath(), SYSTEMSTATE_FILE);
        this.init();
        this.log.end();
    }
    
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.call();
        systemStateFile = null;
    }
}
