/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.transport.platform.joram;

import java.util.concurrent.BlockingQueue;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.jms.Session;

import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.transport.platform.TransportProtocol;
import org.ow2.petals.transport.platform.TransportProtocolListener;
import org.ow2.petals.util.LoggingUtil;

/**
 * Implementation of the JMS message listener which handles async exchanges.
 * When a JMS message is received, the message is unserialized into a JBI
 * exchange, and the exchange is added to the pending exchange Queue. The
 * current thread is waiting a notification on the exchange object. When the
 * thread is awaken, the JMS message is acknowledged.
 * 
 * @author rnaudin
 * 
 */
public class AsyncExchangeMessageListener implements MessageListener {

    /**
     * The logging util
     */
    private final LoggingUtil log;

    /**
     * The Joram Transport Protocol instance
     */
    private final TransportProtocol joramTransporter;

    /**
     * The dispatcher instance
     */
    private final TransportProtocolListener dispatcher;

    /**
     * The name of the related component
     */
    private final String componentName;

    /**
     * A serializer instance
     */
    private final ObjectSerializer serializer = new ObjectSerializer();

    /**
     * The blocking queue where to put the received message
     */
    private final BlockingQueue<MessageExchange> asyncExchangesQueue;

    /**
     * The current thread processing the onMessage() method
     */
    private Thread currentThread;

    /**
     * The Session bound to this message listener
     */
    private Session session;

    /**
     * The running flag
     */
    boolean running;

    /**
     * Creates a new instance of {@link AsyncExchangeMessageListener}
     * 
     * @param log
     * @param dispatcher
     * @param asyncExchangesQueue
     */
    public AsyncExchangeMessageListener(final LoggingUtil log,
            final TransportProtocol joramTransporter, final TransportProtocolListener dispatcher,
            final BlockingQueue<MessageExchange> asyncExchangesQueue, final String componentName,
            final Session session) {
        this.log = log;
        this.joramTransporter = joramTransporter;
        this.dispatcher = dispatcher;
        this.asyncExchangesQueue = asyncExchangesQueue;
        this.componentName = componentName;
        this.session = session;
        this.running = true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jms.MessageListener#onMessage(javax.jms.Message)
     */
    public void onMessage(Message jmsMessage) {
        log.start("session : " + session);
        MessageExchangeImpl exchange = null;

        // When stopping the Joram Transporter or removing a destination,
        // we must not block the thread to let the sessions closing; so we
        // return here
        if (!running) {
            log.debug("Reject a new JMS message during stopping");
            try {
                session.close();
            } catch (JMSException e) {
                log.debug(e);
            }
            log.end();
            return;
        }

        currentThread = Thread.currentThread();

        if (jmsMessage instanceof ObjectMessage) {
            try {
                exchange = serializer.jms2jbi(jmsMessage);
            } catch (JBIException e) {
                log.error("Cannot unserialize the retrieved message for the component '"
                        + componentName + "'");
            }
        } else {
            log.error("The received JMS message is not an ObjectMessage instance");
        }

        // When stopping the Joram Transporter or removing a
        // destination,
        // we must not block the thread to let the sessions closing; so
        // we
        // return here
        if (!running) {
            log.debug("Reject a new JMS message during stopping");
            try {
                session.close();
            } catch (JMSException e) {
                log.debug(e);
            }
            log.end();
            return;
        }

        synchronized (exchange) {
            try {
                asyncExchangesQueue.add(exchange);
                dispatcher.messageAvailable(joramTransporter, componentName);
                if (running) {
                    exchange.wait();
                } else {
                    throw new InterruptedException();
                }
            } catch (InterruptedException e) {
                // The JORAM transporter is stopping
                if (!running) {
                    log.debug("Reject a JMS message during stopping");
                    try {
                        session.close();
                    } catch (JMSException e1) {
                        log.debug(e1);
                    }

                }
                log.end();
                return;
            } finally {
                currentThread = null;
            }
        }

        try {
            jmsMessage.acknowledge();
        } catch (JMSException e) {
            StringBuilder msgBuilder = new StringBuilder();
            msgBuilder.append("Fail to acknowledge the message with id '");
            msgBuilder.append(exchange.getExchangeId());
            msgBuilder.append("' destinated to the component '");
            msgBuilder.append(componentName);
            msgBuilder.append("'");
            log.error(msgBuilder.toString());
        }

        log.end("session : " + session);
    }

    /**
     * Terminate the Message Listener and its bound session. Interrupt the
     * thread waiting to acknowledge a received message.
     * 
     */
    public void terminate() {
        log.start("session : " + session);
        running = false;

        if (currentThread != null) {
            if (currentThread.isAlive()) {
                currentThread.interrupt();
            } else {
                currentThread = null;
            }
        } else {
            try {
                session.close();
            } catch (JMSException e) {
                log.debug(e);
            }
        }

        log.end("session : " + session);
    }
}
