/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.transport.platform.memory;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import javax.jbi.messaging.MessageExchange;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.transport.ProtocolMonitor;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.platform.TransportProtocol;
import org.ow2.petals.transport.platform.TransportProtocolListener;
import org.ow2.petals.util.LoggingUtil;

/**
 * This class provides a local Transporter. The exchanges are stored in Java
 * queues and are never persisted. If a queue exceeds its maximum size, the
 * exchanges are refused by throwing an TransportProtocol exception.
 * 
 * @author Roland NAUDIN - EBM Websourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = TransportProtocol.class))
public class MemoryTransporter implements TransportProtocol {

    /**
     * Name used to identify the local transporter
     */
    public static final String TRANSPORTER_NAME = "memory";

    /**
     * The maximum size of a the Java queue.
     */
    private static final int QUEUE_SIZE = 10000;

    /**
     * The map of the exchanges queues. One queue per installed component.
     */
    private Map<String, BlockingQueue<MessageExchange>> exchangeQueuesMap;

    /**
     * The logger.
     */
    protected LoggingUtil log;

    /**
     * The Fractal component
     */
    @Service
    Component localTransporterComponent;

    /**
     * The fractal exposed interface of the component
     */
    TransportProtocol localTransportProtocol;

    /**
     * Monolog component logger
     */
    @Monolog(name = "logger")
    private Logger logger;

    @Requires(name = "dispatcher", signature = org.ow2.petals.transport.platform.TransportProtocolListener.class)
    private TransportProtocolListener dispatcher;

    private MemoryMonitoring localMonitoring;

    @LifeCycle(on = LifeCycleType.START)
    public void start() throws Exception {
        this.log = new LoggingUtil(this.logger);
        this.log.call();
        this.exchangeQueuesMap = new HashMap<String, BlockingQueue<MessageExchange>>();

        this.localTransportProtocol = (TransportProtocol) this.localTransporterComponent
                .getFcInterface("service");

        this.localMonitoring = new MemoryMonitoring(this);
    }

    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() throws Exception {
        this.log.call();
        for (String componentName : this.exchangeQueuesMap.keySet()) {
            if (!this.exchangeQueuesMap.get(componentName).isEmpty()) {
                this.log
                        .warning("The Memory Transporter contains message destinated to the component '"
                                + componentName + "'. The message are losts.");
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.PetalsLifeCycle#shutdown()
     */
    public void shutdown() throws Exception {
        this.log.call();
        // do nothing
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.jbi.transport.TransportProtocol#addDestination(java
     * .lang.String)
     */
    public void addDestination(final String componentName) throws TransportException {
        this.log.call();
        this.exchangeQueuesMap.put(componentName, new ArrayBlockingQueue<MessageExchange>(
                QUEUE_SIZE));

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.jbi.transport.TransportProtocol#removeDestination
     * (java.lang.String)
     */
    public void removeDestination(final String componentName) throws TransportException {
        this.log.call();
        this.exchangeQueuesMap.remove(componentName);

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.objectweb.petals.jbi.transport.TransportProtocol#receive(java.lang
     * .String)
     */
    public MessageExchangeImpl receive(final String componentName) throws TransportException {
        this.log.start("Receive : " + componentName);

        final MessageExchangeImpl exchange = (MessageExchangeImpl) this.exchangeQueuesMap.get(
                componentName).poll();

        this.log.end();
        return exchange;
    }

    /**
     * Send a message exchange to the given destination. If the message is an
     * response of a synchronous invocation, the
     * <code>dispatcher.syncMessageReceived()</code> is called.
     * 
     * @see org.objectweb.petals.jbi.transport.TransportProtocol#send(org.ow2.petals.jbi.messaging.MessageExchangeImpl,
     *      java.lang.String, java.lang.String, long)
     */
    public void send(final MessageExchangeImpl exchange, final String componentName,
            final String containerName, final String subDomainName, final long timeToLive,
            boolean isSyncResponse) throws TransportException {
        this.log.start();

        if (isSyncResponse) {
            this.dispatcher.syncMessageReceived(exchange);
        } else {
            final BlockingQueue<MessageExchange> exchangesQueue = this.exchangeQueuesMap
                    .get(componentName);

            if (exchangesQueue == null) {
                throw new TransportException("The destination bound to the component '"
                        + componentName + "' is not on the local container");
            }

            if (exchangesQueue.remainingCapacity() == 0) {
                throw new TransportException(
                        "The Memory Transporter has reached its maximum capacity for the target component '"
                                + componentName + "'. The message exchange is rejected");
            }

            exchangesQueue.add(exchange);
            this.dispatcher.messageAvailable(this.localTransportProtocol, componentName);
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.platform.TransportProtocol#getProtocolMonitor()
     */
    public ProtocolMonitor getProtocolMonitor() {
        return this.localMonitoring;
    }

    protected int getQueueSize() {
        return QUEUE_SIZE;
    }

    protected Map<String, BlockingQueue<MessageExchange>> getComponentQueues() {
        return new HashMap<String, BlockingQueue<MessageExchange>>(this.exchangeQueuesMap);
    }

}
