/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.transport.standalone;

import java.util.ArrayList;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeUnit;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.ow2.petals.transport.ProtocolMonitor;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.transport.util.TransporterUtil;
import org.ow2.petals.util.LoggingUtil;

/**
 * @author Roland NAUDIN - EBM WebSourcing
 * 
 */
@FractalComponent
@Provides(interfaces = { @Interface(name = "service", signature = org.ow2.petals.transport.Transporter.class) })
public class StandaloneTransporterImpl implements Transporter, ProtocolMonitor {

    /**
     * Name used to identify the local transporter
     */
    public static final String TRANSPORTER_NAME = "local";

    /**
     * The maximum size of a the Java queue.
     */
    private static final int QUEUE_SIZE = 10000;

    /**
     * The logger.
     */
    private LoggingUtil log;

    /**
     * The map of the exchanges queues. One queue per installed component.
     */
    private Map<String, ArrayBlockingQueue<MessageExchange>> exchangeQueuesMap;

    /**
     * Map of exchanges that are blocked during a synchronous send
     */
    private Map<String, MessageExchangeImpl> pendingExchangesMap;

    /**
     * The list of the threads which are blocking on a poll
     */
    private Vector<Thread> threadsList;

    /**
     * Component started flag
     */
    private boolean componentStarted = false;

    /**
     * Monolog component logger
     */
    @Monolog(name = "logger")
    private Logger logger;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.Transporter#addDestination(java.lang.String)
     */
    public void addDestination(String componentName) throws TransportException {
        this.log.call();
        this.exchangeQueuesMap.put(componentName, new ArrayBlockingQueue<MessageExchange>(
                QUEUE_SIZE));
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.Transporter#removeDestination(java.lang.String)
     */
    public void removeDestination(String componentName) throws TransportException {
        this.log.call();
        this.exchangeQueuesMap.remove(componentName);

    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.Transporter#startTraffic()
     */
    public void startTraffic() {
        // do nothing, as no persistence is provided
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.Transporter#prepareStop()
     */
    public void stopTraffic() {
        this.log.start();
        boolean redo = true;
        this.componentStarted = false;

        /*
         * Awake all the pending 'SYNC_SEND' issuer threads to abort the
         * invocation
         */
        while (redo) {
            try {
                Collection<MessageExchangeImpl> values = this.pendingExchangesMap.values();
                for (MessageExchangeImpl messageExchangeImpl : values) {
                    synchronized (messageExchangeImpl) {
                        messageExchangeImpl.notify();
                    }
                }
                redo = false;
            } catch (ConcurrentModificationException e) {
                // Try again
            }
        }

        redo = true;
        /*
         * Interrupt all the pending receive threads
         */
        while (redo) {
            try {
                for (Thread thread : this.threadsList) {
                    thread.interrupt();
                }
                redo = false;
            } catch (ConcurrentModificationException e) {
                // Try again
            }
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.Transporter#receive(java.lang.String, long)
     */
    public MessageExchangeImpl receive(String componentName, long timeoutMS)
            throws TransportException {
        this.log.start();
        MessageExchangeImpl exchange;

        this.checkTransporterState();

        this.log.debug("Poll a message for the component '" + componentName
                + "' with a timeout of " + timeoutMS);

        if (timeoutMS == 0) {
            // non-blocking call
            exchange = (MessageExchangeImpl) this.exchangeQueuesMap.get(componentName).poll();
        } else {
            this.threadsList.add(Thread.currentThread());
            try {
                if (timeoutMS > 0) {
                    // blocking call with timeout
                    exchange = (MessageExchangeImpl) this.exchangeQueuesMap.get(componentName)
                            .poll(timeoutMS, TimeUnit.MILLISECONDS);
                } else {
                    // blocking call without timeout
                    exchange = (MessageExchangeImpl) this.exchangeQueuesMap.get(componentName)
                            .take();
                }
            } catch (InterruptedException e) {
                // Thread.currentThread().interrupt();
                throw new TransportException("The pending receipt for the component '"
                        + componentName + "' is interrupted", e);
            } finally {
                this.threadsList.remove(Thread.currentThread());
            }
        }

        this.log.end();

        return exchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.Transporter#send(org.ow2.petals.jbi.messaging
     * .exchange.MessageExchangeImpl, java.lang.String, java.lang.String, long)
     */
    public void send(final MessageExchangeImpl exchange, final String componentName,
            final String containerName, final String subDomainName, final long timeToLive)
            throws TransportException {
        this.log.start();

        this.checkTransporterState();

        // clean unused messages
        exchange.cleanMessages();

        ArrayBlockingQueue<MessageExchange> exchangesQueue = this.exchangeQueuesMap
                .get(componentName);

        if (exchangesQueue.remainingCapacity() == 0) {
            throw new TransportException(
                    "The Local Transporter has reached its maximum capacity for the targeted component '"
                            + componentName + "'. The message exchange is rejected");
        }

        if (TransporterUtil.getSyncMode(exchange, true)) {

            // this thread may have processed before the sendSync thread has
            // processed the wait() method
            while (this.pendingExchangesMap.containsKey(exchange.getExchangeId())) {
                synchronized (exchange) {
                    exchange.notify();
                }
                try {
                    Thread.sleep(1);
                } catch (InterruptedException e) {
                    throw new TransportException(e);
                }
            }
        } else {
            exchangesQueue.add(exchange);
        }

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.Transporter#sendSync(org.ow2.petals.jbi.messaging
     * .exchange.MessageExchangeImpl, java.lang.String, java.lang.String, long)
     */
    public MessageExchangeImpl sendSync(final MessageExchangeImpl exchange,
            final String componentName, final String containerName, final String subDomainName,
            final long timeToLive) throws TransportException {
        this.log.start();

        this.checkTransporterState();

        // clean unused messages
        exchange.cleanMessages();

        // synchronous send
        try {
            TransporterUtil.setSendSyncProperties(exchange);
        } catch (MessagingException e) {
            throw new TransportException(e);
        }

        // Reference this exchange in the Map of waitingExchanges.
        // It will be used to wake up the exchange when the response
        // is arrived
        this.pendingExchangesMap.put(exchange.getExchangeId(), exchange);

        ArrayBlockingQueue<MessageExchange> exchangesQueue = this.exchangeQueuesMap
                .get(componentName);

        if (exchangesQueue.remainingCapacity() == 0) {
            throw new TransportException(
                    "The Local Transporter has reached its maximum capacity for the targeted component '"
                            + componentName + "'. The message exchange is rejected");
        }
        
        try {
            if (TransporterUtil.getSyncMode(exchange, true)) {
                // in this case the current send sync is a response of a
                // previous send sync
                synchronized (exchange) {
                    exchange.notify();
                }
            } else {
                exchangesQueue.add(exchange);
            }

            // wait for a response or a timeout reached
            synchronized (exchange) {
                exchange.wait(timeToLive);
            }

        } catch (InterruptedException e) {
            throw new TransportException(
                    "Failed to receive the response message of a synchronous send", e);
        } finally {
            // try to remove the exchange if it has not been already consumed by
            // the target component
            if (!TransporterUtil.updateSyncProperties(exchange)) {
                exchangesQueue.remove(exchange);
                this.pendingExchangesMap.remove(exchange.getExchangeId());
            }
        }

        // If the synchronous property is set,
        // it means that the timeout has been reached or the transporter is
        // stopping
        if (TransporterUtil.getSyncMode(exchange, false)) {
            this.checkTransporterState();

            this.log.warning("Failed to send synchronously the exchange: "
                    + exchange.getExchangeId() + ". Timeout occured");

            this.log.end();
            return null;
        }

        this.log.end();

        return exchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.PetalsLifeCycle#shutdown()
     */
    public void shutdown() throws Exception {
        // nothing to do

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.transport.Transporter#ProtocolMonitor(java.lang.String)
     */
    public org.ow2.petals.transport.ProtocolMonitor getProtocolMonitor(String protocolName)
            throws TransportException {
        return this;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.Transporter#getTransporters()
     */
    public List<String> getTransporters() {
        List<String> transportersList = new ArrayList<String>();
        transportersList.add(TRANSPORTER_NAME);

        return transportersList;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.ProtocolMonitor#getQueueMaxSize()
     */
    public int getQueueMaxSize() throws TransportException {
        return QUEUE_SIZE;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.transport.ProtocolMonitor#getQueueSizes()
     */
    public Map<String, Integer> getQueueSizes() throws TransportException {
        Map<String, Integer> result = new HashMap<String, Integer>();
        for (String componentName : this.exchangeQueuesMap.keySet()) {
            result.put(componentName, this.exchangeQueuesMap.get(componentName).size());
        }

        return result;
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws Exception {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        this.exchangeQueuesMap = new Hashtable<String, ArrayBlockingQueue<MessageExchange>>();
        this.pendingExchangesMap = new Hashtable<String, MessageExchangeImpl>();
        this.threadsList = new Vector<Thread>();
        this.componentStarted = true;

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws Exception {
        this.log.call();
    }

    /**
     * Check if the transporter is running, if not, raise an exception.
     * 
     * @throws TransportException
     */
    private void checkTransporterState() throws TransportException {
        this.log.call();

        if (!this.componentStarted) {
            Thread.currentThread().interrupt();
            throw new TransportException("The Transporter is stopped", new InterruptedException());
        }
    }

}
